"""Tests the `vertex.call` module."""

import sys
from unittest.mock import MagicMock, patch

from mirascope.core.vertex import _utils
from mirascope.core.vertex.call_response import VertexCallResponse
from mirascope.core.vertex.call_response_chunk import VertexCallResponseChunk
from mirascope.core.vertex.stream import VertexStream
from mirascope.core.vertex.tool import VertexTool


def test_vertex_call() -> None:
    """Tests the `vertex_call` decorator."""

    if "mirascope.core.vertex._call" in sys.modules:
        del sys.modules["mirascope.core.vertex._call"]

    with patch(
        "mirascope.core.base.call_factory", new_callable=MagicMock
    ) as mock_call_factory:
        import mirascope.core.vertex._call  # noqa: F401

        mock_call_factory.assert_called_once_with(
            TCallResponse=VertexCallResponse,
            TCallResponseChunk=VertexCallResponseChunk,
            TToolType=VertexTool,
            TStream=VertexStream,
            default_call_params={},
            setup_call=_utils.setup_call,
            get_json_output=_utils.get_json_output,
            handle_stream=_utils.handle_stream,
            handle_stream_async=_utils.handle_stream_async,
        )



import asyncio
import wave
from collections.abc import AsyncGenerator, Awaitable, Callable
from io import BytesIO
from queue import Empty, Queue
from typing import Any

import numpy
import numpy as np
import sounddevice as sd

from ._utils._audio import CHANNELS, FRAME_RATE


async def async_input(prompt: str) -> str:
    return await asyncio.to_thread(input, prompt)


def _create_buffer_from_recording(recording: numpy.ndarray) -> BytesIO:
    buffer = BytesIO()
    with wave.open(buffer, "wb") as wf:
        wf.setnchannels(1)
        wf.setsampwidth(2)  # int16 is 2 bytes
        wf.setframerate(FRAME_RATE)
        wf.writeframes(recording.tobytes())
    buffer.seek(0)
    return buffer


async def record_as_stream(
    custom_blocking_event: Callable[..., Awaitable[...]] | None = None,
) -> AsyncGenerator[BytesIO, None]:
    event = asyncio.Event()
    data_queue = Queue()

    def callback(indata: np.ndarray, *args: Any) -> None:  # noqa: ANN401
        if event.is_set():
            raise sd.CallbackStop
        data_queue.put(indata.copy())

    stream = sd.InputStream(
        samplerate=FRAME_RATE, channels=CHANNELS, dtype="int16", callback=callback
    )

    async def blocking_event() -> None:
        if custom_blocking_event:
            await custom_blocking_event()
        else:
            while True:
                await asyncio.sleep(0.01)
        event.set()

    with stream:
        asyncio.create_task(blocking_event())
        accumulated_data = []
        accumulated_bytes = 0
        chunk_size_bytes = 32 * 1024  # 32KB
        while not event.is_set() or not data_queue.empty():
            try:
                indata_ = data_queue.get_nowait()
                accumulated_data.append(indata_)
                accumulated_bytes += indata_.nbytes

                if accumulated_bytes >= chunk_size_bytes:
                    data_to_yield = np.concatenate(accumulated_data, axis=0)

                    buffer = _create_buffer_from_recording(data_to_yield)
                    yield buffer

                    accumulated_data = []
                    accumulated_bytes = 0
            except Empty:
                await asyncio.sleep(0.01)

    while not data_queue.empty():
        yield data_queue.get_nowait()


async def record(custom_blocking_event: Callable[..., Awaitable[...]]) -> BytesIO:
    recording = []
    event = asyncio.Event()

    def callback(indata: numpy.ndarray, *args: Any) -> None:  # noqa: ANN401
        if event.is_set():
            raise sd.CallbackStop
        recording.append(indata.copy())

    stream = sd.InputStream(
        samplerate=FRAME_RATE, channels=CHANNELS, dtype="int16", callback=callback
    )

    with stream:
        await custom_blocking_event()
        event.set()

    recording = np.concatenate(recording, axis=0)

    return _create_buffer_from_recording(recording)



from mirascope.core import BaseMessageParam, openai


@openai.call("gpt-4o-mini")
def summarize(text: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Summarize this text: {text}")]


@openai.call("gpt-4o-mini")
def summarize_and_translate(text: str, language: str) -> list[BaseMessageParam]:
    summary = summarize(text)
    return [
        BaseMessageParam(
            role="user",
            content=f"Translate this text to {language}: {summary.content}",
        )
    ]


response = summarize_and_translate("Long English text here...", "french")
print(response.content)



from mirascope.core import openai, prompt_template


@openai.call("gpt-4o-mini", call_params={"max_tokens": 512})
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


response = recommend_book("fantasy")
print(response.content)



from __future__ import annotations

from collections.abc import Sequence
from io import FileIO
from typing import cast
from wave import Wave_read

from pydantic import BaseModel
from typing_extensions import TypeIs

from .._utils._get_image_type import get_image_type
from ..message_param import (
    AudioPart,
    BaseMessageParam,
    CacheControlPart,
    ImagePart,
    TextPart,
)
from ..types import AudioSegment, Image, has_pil_module, has_pydub_module

SAMPLE_WIDTH = 2
FRAME_RATE = 24000
CHANNELS = 1


def _convert_message_sequence_part_to_content_part(
    message_sequence_part: str
    | Image.Image
    | TextPart
    | ImagePart
    | AudioPart
    | AudioSegment
    | Wave_read,
) -> TextPart | ImagePart | AudioPart | CacheControlPart:
    if isinstance(message_sequence_part, str):
        return TextPart(text=message_sequence_part, type="text")
    elif isinstance(
        message_sequence_part, TextPart | ImagePart | AudioPart | CacheControlPart
    ):
        return message_sequence_part
    elif has_pil_module and isinstance(message_sequence_part, Image.Image):
        image = message_sequence_part.tobytes()
        return ImagePart(
            type="image",
            media_type=f"image/{get_image_type(image)}",
            image=image,
            detail=None,
        )
    elif has_pydub_module and isinstance(message_sequence_part, AudioSegment):
        return AudioPart(
            type="audio",
            media_type="audio/wav",
            audio=cast(
                FileIO,
                message_sequence_part.set_frame_rate(FRAME_RATE)
                .set_channels(CHANNELS)
                .set_sample_width(SAMPLE_WIDTH)
                .export(format="wav"),
            ).read(),
        )
    elif isinstance(message_sequence_part, Wave_read):
        return AudioPart(
            type="audio",
            media_type="audio/wav",
            audio=message_sequence_part.readframes(-1),
        )
    else:
        raise ValueError(f"Invalid message sequence type: {message_sequence_part}")


def convert_message_content_to_message_param_content(
    message_sequence: Sequence[
        str | TextPart | ImagePart | Image.Image | AudioPart | AudioSegment | Wave_read
    ],
) -> list[TextPart | ImagePart | AudioPart | CacheControlPart] | str:
    if isinstance(message_sequence, str):
        return message_sequence
    return [
        _convert_message_sequence_part_to_content_part(message)
        for message in message_sequence
    ]


def _is_base_message_params(
    value: object,
) -> TypeIs[list[BaseMessageParam]]:
    # Note: we also need to catch the original provider message parameters here
    return isinstance(value, list) and all(
        isinstance(v, BaseMessageParam | dict | BaseModel) for v in value
    )


def convert_messages_to_message_params(
    messages: str
    | Sequence[
        str | TextPart | ImagePart | Image.Image | AudioPart | AudioSegment | Wave_read
    ]
    | list[BaseMessageParam]
    | BaseMessageParam,
    role: str = "user",
) -> list[BaseMessageParam]:
    if isinstance(messages, BaseMessageParam):
        return [messages]
    elif _is_base_message_params(messages):
        return messages
    elif isinstance(messages, str | Sequence):
        return [
            BaseMessageParam(
                content=convert_message_content_to_message_param_content(messages),
                role=role,
            )
        ]
    else:
        raise ValueError(f"Invalid messages type: {messages}")



from mirascope.core import BaseMessageParam, bedrock


@bedrock.call(
    "anthropic.claude-3-haiku-20240307-v1:0",
    call_params={"inferenceConfig": {"maxTokens": 512}},
)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


response = recommend_book("fantasy")
print(response.content)



"""The Mirascope LiteLLM Module."""

from typing import TypeAlias

from ..openai import (
    OpenAICallResponse,
    OpenAICallResponseChunk,
    OpenAIDynamicConfig,
    OpenAIMessageParam,
)
from ._call import litellm_call
from ._call import litellm_call as call

LiteLLMMessageParam: TypeAlias = OpenAIMessageParam

__all__ = ["call", "LiteLLMMessageParam", "litellm_call"]



from google.generativeai import GenerativeModel
from mirascope.core import BaseMessageParam, gemini


@gemini.call("", client=GenerativeModel(model_name="gemini-1.5-flash"))
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



from azure.ai.inference.aio import ChatCompletionsClient as AsyncChatCompletionsClient
from azure.core.credentials import AzureKeyCredential
from mirascope.core import BaseMessageParam, azure


@azure.call(
    "gpt-4o-mini",
    client=AsyncChatCompletionsClient(
        endpoint="your-endpoint", credential=AzureKeyCredential("your-credentials")
    ),
)
async def recommend_book_async(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



import asyncio
from io import BytesIO

from typing import AsyncGenerator, Any

from pydub import AudioSegment
from pydub.playback import play

from mirascope.beta.openai import Realtime, record_as_stream

app = Realtime(
    "gpt-4o-realtime-preview-2024-10-01",
)


@app.receiver("audio")
async def receive_audio(response: AudioSegment, context: dict[str, Any]) -> None:
    play(response)


@app.receiver("audio_transcript")
async def receive_audio_transcript(response: str, context: dict[str, Any]) -> None:
    print(f"AI(audio_transcript): {response}")


@app.sender()
async def send_audio_as_stream(
    context: dict[str, Any],
) -> AsyncGenerator[BytesIO, None]:
    print("Sending audio...")
    async for stream in record_as_stream():
        yield stream


asyncio.run(app.run())



from anthropic import AnthropicError
from mirascope.core import Messages, anthropic


@anthropic.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


try:
    response = recommend_book("fantasy")
    print(response.content)
except AnthropicError as e:
    print(f"Error: {str(e)}")



from mirascope.core import bedrock, prompt_template


@bedrock.call(
    "anthropic.claude-3-haiku-20240307-v1:0",
    call_params={"inferenceConfig": {"maxTokens": 512}},
)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import azure, prompt_template


@azure.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    response = recommend_book("fantasy")
    print(response.content)
except Exception as e:
    print(f"Error: {str(e)}")



from mirascope.core import litellm


@litellm.call("gpt-4o-mini", call_params={"max_tokens": 512})
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("fantasy")
print(response.content)



def exact_match_eval(output: str, expected: list[str]) -> bool:
    return all(phrase in output for phrase in expected)


# Example usage
output = "The capital of France is Paris, and it's known for the Eiffel Tower."
expected = ["capital of France", "Paris", "Eiffel Tower"]
result = exact_match_eval(output, expected)
print(result)  # Output: True



from mirascope.core import azure, prompt_template
from mirascope.integrations.otel import with_hyperdx


@with_hyperdx()
@azure.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



from mirascope.core import (
    BaseDynamicConfig,
    BaseToolKit,
    litellm,
    prompt_template,
    toolkit_tool,
)


class BookTools(BaseToolKit):
    __namespace__ = "book_tools"

    reading_level: str

    @toolkit_tool
    def suggest_author(self, author: str) -> str:
        """Suggests an author for the user to read based on their reading level.

        User reading level: {self.reading_level}
        """
        return f"I would suggest you read some books by {author}"


@litellm.call("gpt-4o-mini")
@prompt_template("What {genre} author should I read?")
def recommend_author(genre: str, reading_level: str) -> BaseDynamicConfig:
    toolkit = BookTools(reading_level=reading_level)
    return {"tools": toolkit.create_tools()}


response = recommend_author("fantasy", "beginner")
if tool := response.tool:
    print(tool.call())
    # Output: I would suggest you read some books by J.K. Rowling

response = recommend_author("fantasy", "advanced")
if tool := response.tool:
    print(tool.call())
    # Output: I would suggest you read some books by Brandon Sanderson



from mirascope.core import openai, prompt_template
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@openai.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



from anthropic import AnthropicError
from mirascope.core import anthropic, prompt_template


@anthropic.call("claude-3-5-sonnet-20240620", stream=True)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except AnthropicError as e:
    print(f"Error: {str(e)}")



from mirascope.core import BaseMessageParam, anthropic, openai, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


# OpenAI
openai_model = "gpt-4o-mini"
openai_recommend_book = openai.call(openai_model)(recommend_book_prompt)
openai_response = openai_recommend_book("fantasy")
print(openai_response.content)

# Anthropic
anthropic_model = "claude-3-5-sonnet-20240620"
anthropic_recommend_book = anthropic.call(anthropic_model)(recommend_book_prompt)
anthropic_response = anthropic_recommend_book("fantasy")
print(anthropic_response.content)



"""This module contains the setup_call function for Bedrock tools."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine, Generator
from functools import wraps
from typing import Any, ParamSpec, cast, overload

from aiobotocore.session import AioSession, get_session
from boto3.session import Session
from mypy_boto3_bedrock_runtime import BedrockRuntimeClient
from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef,
    ConverseStreamResponseTypeDef,
    ToolChoiceTypeDef,
    ToolConfigurationTypeDef,
)
from types_aiobotocore_bedrock_runtime import (
    BedrockRuntimeClient as AsyncBedrockRuntimeClient,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef as AsyncConverseResponseTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseStreamResponseTypeDef as AsyncConverseStreamResponseTypeDef,
)

from ... import BaseMessageParam
from ...base import BaseTool, _utils
from ...base._utils import (
    AsyncCreateFn,
    CreateFn,
    fn_is_async,
    get_async_create_fn,
    get_create_fn,
)
from .._types import (
    AsyncStreamOutputChunk,
    InternalBedrockMessageParam,
    StreamOutputChunk,
)
from ..call_kwargs import BedrockCallKwargs
from ..call_params import BedrockCallParams
from ..dynamic_config import BedrockDynamicConfig
from ..tool import BedrockTool
from ._convert_message_params import convert_message_params

_P = ParamSpec("_P")


def _extract_sync_stream_fn(
    fn: Callable[_P, ConverseStreamResponseTypeDef], model: str
) -> Callable[_P, Generator[StreamOutputChunk, None, None]]:
    @wraps(fn)
    def _inner(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> Generator[StreamOutputChunk, None, None]:
        response = fn(*args, **kwargs)
        for chunk in response["stream"]:
            yield StreamOutputChunk(
                responseMetadata=response["ResponseMetadata"], model=model, **chunk
            )

    return _inner


def _extract_async_stream_fn(
    fn: Callable[_P, Coroutine[Any, Any, AsyncConverseStreamResponseTypeDef]],
    model: str,
) -> Callable[_P, AsyncGenerator[AsyncStreamOutputChunk, None]]:
    @wraps(fn)
    async def _inner(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> AsyncGenerator[AsyncStreamOutputChunk, None]:
        response = await fn(*args, **kwargs)
        async for chunk in response["stream"]:
            yield AsyncStreamOutputChunk(
                responseMetadata=response["ResponseMetadata"], model=model, **chunk
            )

    return _inner


async def _get_async_client(session: AioSession) -> AsyncBedrockRuntimeClient:
    async with session.create_client("bedrock-runtime") as client:
        return client


@overload
def setup_call(
    *,
    model: str,
    client: AsyncBedrockRuntimeClient | None,
    fn: Callable[..., Awaitable[BedrockDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: BedrockDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: BedrockCallParams,
    extract: bool,
) -> tuple[
    AsyncCreateFn[AsyncConverseResponseTypeDef, AsyncStreamOutputChunk],
    str | None,
    list[InternalBedrockMessageParam],
    list[type[BedrockTool]] | None,
    BedrockCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: BedrockRuntimeClient | None,
    fn: Callable[..., BedrockDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: BedrockDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: BedrockCallParams,
    extract: bool,
) -> tuple[
    CreateFn[ConverseResponseTypeDef, StreamOutputChunk],
    str | None,
    list[InternalBedrockMessageParam],
    list[type[BedrockTool]] | None,
    BedrockCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: BedrockRuntimeClient | AsyncBedrockRuntimeClient | None,
    fn: Callable[..., BedrockDynamicConfig]
    | Callable[..., Awaitable[BedrockDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: BedrockDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: BedrockCallParams,
    extract: bool,
) -> tuple[
    AsyncCreateFn[AsyncConverseResponseTypeDef, AsyncStreamOutputChunk]
    | CreateFn[ConverseResponseTypeDef, StreamOutputChunk],
    str | None,
    list[InternalBedrockMessageParam],
    list[type[BedrockTool]] | None,
    BedrockCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, BedrockTool, call_params
    )
    call_kwargs = cast(BedrockCallKwargs, base_call_kwargs)
    messages = cast(list[InternalBedrockMessageParam | BaseMessageParam], messages)
    messages = convert_message_params(messages)
    if messages[0]["role"] == "system":
        call_kwargs["system"] = [
            {"text": text}
            for c in messages.pop(0)["content"]
            if (text := c.get("text"))
        ]

    call_kwargs_tools = call_kwargs.pop("tools", None)
    if json_mode:
        json_mode_content = _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        if "text" in messages[-1]["content"][-1]:
            messages[-1]["content"][-1]["text"] += json_mode_content
        else:
            messages[-1]["content"] = [
                *messages[-1]["content"],
                {"text": json_mode_content},
            ]

    else:
        if call_kwargs_tools:
            call_kwargs["toolConfig"] = cast(
                ToolConfigurationTypeDef, {"tools": call_kwargs_tools}
            )
        if extract:
            assert tool_types, "At least one tool must be provided for extraction."
            if "toolConfig" in call_kwargs:
                call_kwargs["toolConfig"]["toolChoice"] = cast(
                    ToolChoiceTypeDef, {"type": "tool", "name": tool_types[0]._name()}
                )

    call_kwargs |= cast(BedrockCallKwargs, {"modelId": model, "messages": messages})

    if client is None:
        if fn_is_async(fn):
            session = get_session()
            client = asyncio.run(_get_async_client(session))
        else:
            session = Session()
            client = session.client("bedrock-runtime")

    create = (
        get_async_create_fn(
            client.converse, _extract_async_stream_fn(client.converse_stream, model)
        )
        if isinstance(client, AsyncBedrockRuntimeClient)
        else get_create_fn(
            client.converse, _extract_sync_stream_fn(client.converse_stream, model)
        )
    )
    return create, prompt_template, messages, tool_types, call_kwargs



from google.generativeai import GenerativeModel
from mirascope.core import gemini, prompt_template


@gemini.call("", client=GenerativeModel(model_name="gemini-1.5-flash"))
@prompt_template("Recommend a {genre} book")
async def recommend_book_async(genre: str): ...



import asyncio

from mirascope.core import BaseTool, litellm


class FormatBook(BaseTool):
    title: str
    author: str

    async def call(self) -> str:
        # Simulating an async API call
        await asyncio.sleep(1)
        return f"{self.title} by {self.author}"


@litellm.call(model="gpt-4o-mini", tools=[FormatBook])
async def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


async def main():
    response = await recommend_book("fantasy")
    if tool := response.tool:
        if isinstance(tool, FormatBook):
            output = await tool.call()
            print(output)
    else:
        print(response.content)


asyncio.run(main())



from mirascope.core import BaseTool, anthropic
from mirascope.core.anthropic import AnthropicToolConfig


class CachedTool(BaseTool):
    """This is an example of a cached tool."""

    tool_config = AnthropicToolConfig(cache_control={"type": "ephemeral"})

    def call(self) -> str:
        return "Example tool"


@anthropic.call(
    "claude-3-5-sonnet-20240620",
    tools=[CachedTool],
    call_params={
        "max_tokens": 1024,
        "extra_headers": {"anthropic-beta": "prompt-caching-2024-07-31"},
    },
)
def cached_tool_call() -> str:
    return "An example call with a cached tool"



from mirascope.core import mistral
from mistralai import models


@mistral.call("mistral-large-latest", stream=True)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except models.HTTPValidationError as e:  # pyright: ignore [reportAttributeAccessIssue]
    # handle e.data: models.HTTPValidationErrorData
    raise (e)
except models.SDKError as e:  # pyright: ignore [reportAttributeAccessIssue]
    # handle exception
    raise (e)



from azure.ai.inference import ChatCompletionsClient
from azure.core.credentials import AzureKeyCredential
from mirascope.core import Messages, azure


@azure.call(
    "gpt-4o-mini",
    client=ChatCompletionsClient(
        endpoint="your-endpoint", credential=AzureKeyCredential("your-credentials")
    ),
)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



"""Tests the `mistral._utils.calculate_cost` function."""

from mirascope.core.mistral._utils._calculate_cost import calculate_cost


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function."""
    assert calculate_cost(None, None, model="mistral-large-latest") is None
    assert calculate_cost(1, 1, model="unknown") is None
    assert calculate_cost(1, 1, model="mistral-large-latest") == 1.2e-5



from mirascope.core import Messages, openai
from openai import OpenAI


@openai.call("gpt-4o-mini", client=OpenAI())
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



from mirascope.core import BaseDynamicConfig, Messages, anthropic


@anthropic.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> BaseDynamicConfig:
    return {
        "messages": [Messages.User(f"Recommend a {genre} book")],
        "call_params": {"max_tokens": 512},
        "metadata": {"tags": {"version:0001"}},
    }


response = recommend_book("fantasy")
print(response.content)



"""This module contains the type definition for the Gemini call keyword arguments."""

from google.generativeai.types import ContentDict, Tool

from ..base import BaseCallKwargs
from .call_params import GeminiCallParams


class GeminiCallKwargs(GeminiCallParams, BaseCallKwargs[Tool]):
    contents: list[ContentDict]



"""Utility for converting `BaseMessageParam` to `Content`"""

from vertexai.generative_models import Content, Image, Part

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | Content],
) -> list[Content]:
    converted_message_params = []
    for message_param in message_params:
        if isinstance(message_param, Content):
            converted_message_params.append(message_param)
        elif (role := message_param.role) == "system":
            content = message_param.content
            if not isinstance(content, str):
                raise ValueError(
                    "System message content must be a single text string."
                )  # pragma: no cover
            converted_message_params += [
                Content(
                    role="user",
                    parts=[
                        Part.from_text(content) if isinstance(content, str) else content
                    ],
                ),
                Content(
                    role="model",
                    parts=[Part.from_text("Ok! I will adhere to this system message.")],
                ),
            ]
        elif isinstance((content := message_param.content), str):
            converted_message_params.append(
                Content(role="user", parts=[Part.from_text(content)])
            )
        else:
            converted_content: list[Part] = []
            for part in content:
                if part.type == "text":
                    converted_content.append(Part.from_text(part.text))
                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/webp",
                        "image/heic",
                        "image/heif",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. "
                            "Vertex currently only supports JPEG, PNG, WebP, HEIC, "
                            "and HEIF images."
                        )
                    image = Image.from_bytes(part.image)
                    converted_content.append(Part.from_image(image))
                elif part.type == "audio":
                    if part.media_type not in [
                        "audio/wav",
                        "audio/mp3",
                        "audio/aiff",
                        "audio/aac",
                        "audio/ogg",
                        "audio/flac",
                    ]:
                        raise ValueError(
                            f"Unsupported audio media type: {part.media_type}. "
                            "Vertex currently only supports WAV, MP3, AIFF, AAC, OGG, "
                            "and FLAC audio file types."
                        )
                    converted_content.append(
                        Part.from_data(mime_type=part.media_type, data=part.audio)
                    )
                else:
                    raise ValueError(
                        "Vertex currently only supports text, image, and audio parts. "
                        f"Part provided: {part.type}"
                    )
            converted_message_params.append(Content(role=role, parts=converted_content))
    return converted_message_params



"""Tests the `azure.stream` module."""

from datetime import datetime

import pytest
from azure.ai.inference.models import (
    ChatChoice,
    ChatCompletions,
    ChatCompletionsToolCall,
    ChatRequestMessage,
    ChatResponseMessage,
    CompletionsFinishReason,
    CompletionsUsage,
    FunctionCall,
    StreamingChatChoiceUpdate,
    StreamingChatCompletionsUpdate,
    StreamingChatResponseMessageUpdate,
    StreamingChatResponseToolCallUpdate,
    UserMessage,
)

from mirascope.core.azure.call_response import AzureCallResponse
from mirascope.core.azure.call_response_chunk import AzureCallResponseChunk
from mirascope.core.azure.stream import AzureStream
from mirascope.core.azure.tool import AzureTool


@pytest.fixture
def mock_datetime_now(monkeypatch):
    class MockDatetime:
        @classmethod
        def now(cls):
            return datetime.fromtimestamp(0)

    monkeypatch.setattr("datetime.datetime", MockDatetime)
    yield

    monkeypatch.undo()


def test_azure_stream() -> None:
    """Tests the `AzureStream` class."""
    assert AzureStream._provider == "azure"

    class FormatBook(AzureTool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> None:
            """Dummy call."""

    tool_call_delta = StreamingChatResponseToolCallUpdate(
        id="id",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
    )
    created = datetime.fromtimestamp(0)
    usage = CompletionsUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    chunks = [
        StreamingChatCompletionsUpdate(
            id="id",
            choices=[
                StreamingChatChoiceUpdate(
                    delta=StreamingChatResponseMessageUpdate(
                        content="content", tool_calls=None
                    ),
                    index=0,
                    finish_reason=CompletionsFinishReason("stop"),
                )
            ],
            created=created,
            model="gpt-4o",
            usage=CompletionsUsage(
                completion_tokens=0, prompt_tokens=0, total_tokens=0
            ),
        ),
        StreamingChatCompletionsUpdate(
            id="id",
            choices=[
                StreamingChatChoiceUpdate(
                    delta=StreamingChatResponseMessageUpdate(
                        content=None,
                        tool_calls=[tool_call_delta],
                    ),
                    index=0,
                    finish_reason=CompletionsFinishReason("stop"),
                )
            ],
            created=created,
            model="gpt-4o",
            usage=usage,
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = AzureCallResponseChunk(chunk=chunk)
            if tool_calls := call_response_chunk.chunk.choices[0].delta.tool_calls:
                assert tool_calls[0].function
                tool_call = ChatCompletionsToolCall(
                    id="id",
                    function=FunctionCall(**tool_calls[0].function),
                )
                yield (
                    call_response_chunk,
                    FormatBook.from_tool_call(tool_call),
                )
            else:
                yield call_response_chunk, None

    stream = AzureStream(
        stream=generator(),
        metadata={},
        tool_types=[FormatBook],
        call_response_type=AzureCallResponse,
        model="gpt-4o",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[ChatRequestMessage({"role": "user", "content": "content"})],
        call_params={},
        call_kwargs={},
    )

    with pytest.raises(
        ValueError, match="No stream response, check if the stream has been consumed."
    ):
        stream.construct_call_response()

    assert stream.cost is None
    for _ in stream:
        pass
    assert stream.cost is None

    format_book = FormatBook.from_tool_call(
        ChatCompletionsToolCall(
            id="id",
            function=FunctionCall(
                arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
                name="FormatBook",
            ),
        )
    )
    assert stream.message_param == {
        "role": "assistant",
        "content": "content",
        "tool_calls": [format_book.tool_call],
    }


def test_construct_call_response(mock_datetime_now) -> None:
    """Tests the `AzureStream.construct_call_response` method."""

    class FormatBook(AzureTool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> None:
            """Dummy call."""

    tool_call = StreamingChatResponseToolCallUpdate(
        id="id",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
    )
    usage = CompletionsUsage(prompt_tokens=1, completion_tokens=1, total_tokens=2)
    chunks = [
        StreamingChatCompletionsUpdate(
            id="id",
            model="gpt-4o",
            choices=[
                StreamingChatChoiceUpdate(
                    delta=StreamingChatResponseMessageUpdate(
                        content="content", tool_calls=None
                    ),
                    index=0,
                    finish_reason=CompletionsFinishReason("stop"),
                )
            ],
            created=datetime.fromtimestamp(0),
            usage=CompletionsUsage(
                completion_tokens=0, prompt_tokens=0, total_tokens=0
            ),
        ),
        StreamingChatCompletionsUpdate(
            id="id",
            model="gpt-4o",
            choices=[
                StreamingChatChoiceUpdate(
                    delta=StreamingChatResponseMessageUpdate(
                        content=None,
                        tool_calls=[tool_call],
                    ),
                    index=0,
                    finish_reason=CompletionsFinishReason("stop"),
                )
            ],
            created=datetime.fromtimestamp(0),
            usage=usage,
        ),
    ]

    tool_call = None

    def generator():
        nonlocal tool_call
        for chunk in chunks:
            call_response_chunk = AzureCallResponseChunk(chunk=chunk)
            if tool_calls := call_response_chunk.chunk.choices[0].delta.tool_calls:
                assert tool_calls[0].function
                tool_call = ChatCompletionsToolCall(
                    id="id",
                    function=FunctionCall(**tool_calls[0].function),
                )
                yield (
                    call_response_chunk,
                    FormatBook.from_tool_call(tool_call),
                )
            else:
                yield call_response_chunk, None

    stream = AzureStream(
        stream=generator(),
        metadata={},
        tool_types=[FormatBook],
        call_response_type=AzureCallResponse,
        model="gpt-4o",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[UserMessage(content="content")],
        call_params={},
        call_kwargs={},
    )

    for _ in stream:
        pass

    tool_call = ChatCompletionsToolCall(
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        id="id",
    )
    completion = ChatCompletions(
        choices=[
            ChatChoice(
                finish_reason="stop",
                index=0,
                message=ChatResponseMessage(
                    role="assistant", content="content", tool_calls=[tool_call]
                ),
            )
        ],
        created=datetime.fromtimestamp(0),
        model="gpt-4o",
        usage=usage,
        id="id",
    )
    call_response = AzureCallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    constructed_call_response = stream.construct_call_response()
    assert dict(constructed_call_response.response) == dict(call_response.response)


def test_construct_call_response_no_usage() -> None:
    """Tests the `AzureStream.construct_call_response` method with no usage."""
    chunks = [
        StreamingChatCompletionsUpdate(
            id="id",
            choices=[
                StreamingChatChoiceUpdate(
                    delta=StreamingChatResponseMessageUpdate(
                        content="content", tool_calls=None
                    ),
                    index=0,
                    finish_reason=CompletionsFinishReason("stop"),
                )
            ],
            created=datetime.fromtimestamp(0),
            model="gpt-4o",
            usage=CompletionsUsage(
                completion_tokens=0, prompt_tokens=0, total_tokens=0
            ),
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = AzureCallResponseChunk(chunk=chunk)
            yield call_response_chunk, None

    stream = AzureStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=AzureCallResponse,
        model="gpt-4o",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[UserMessage(content="content")],
        call_params={},
        call_kwargs={},
    )
    for _ in stream:
        pass
    constructed_call_response = stream.construct_call_response()
    assert constructed_call_response.usage == {
        "completion_tokens": 0,
        "prompt_tokens": 0,
        "total_tokens": 0,
    }



from mirascope.core import BaseMessageParam, cohere


def parse_recommendation(response: cohere.CohereCallResponse) -> tuple[str, str]:
    title, author = response.content.split(" by ")
    return (title, author)


@cohere.call("command-r-plus", output_parser=parse_recommendation)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user",
            content=f"Recommend a {genre} book. Output only Title by Author",
        )
    ]


print(recommend_book("fantasy"))
# Output: ('"The Name of the Wind"', 'Patrick Rothfuss')



from mirascope.core import prompt_template


@prompt_template(
    """
    Recommend a {genre} book.
    Output in the format Title by Author.
    """
)
def recommend_book_prompt(genre: str): ...


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='system', content='Recommend a fantasy book.\nOutput in the format Title by Author.')]



from mirascope.core import mistral
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@mistral.call("mistral-large-latest")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



import io


from pydub.playback import play
from pydub import AudioSegment

from mirascope.core import openai, BaseMessageParam

SAMPLE_WIDTH = 2
FRAME_RATE = 24000
CHANNELS = 1


@openai.call(
    "gpt-4o-audio-preview",
    call_params={
        "audio": {"voice": "alloy", "format": "pcm16"},
        "modalities": ["text", "audio"],
    },
    stream=True,
)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


audio_chunk = b""
audio_transcript_chunk = ""

stream = recommend_book("fantasy")
for chunk, _ in stream:
    if chunk.audio:
        audio_chunk += chunk.audio
    if chunk.audio_transcript:
        audio_transcript_chunk += chunk.audio_transcript

print(audio_transcript_chunk)


audio_segment = AudioSegment.from_raw(
    io.BytesIO(audio_chunk),
    sample_width=SAMPLE_WIDTH,
    frame_rate=FRAME_RATE,
    channels=CHANNELS,
)
play(audio_segment)



from cohere.errors import BadRequestError
from mirascope.core import cohere, prompt_template


@cohere.call("command-r-plus", stream=True)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except BadRequestError as e:
    print(f"Error: {str(e)}")



from mirascope.core import BaseDynamicConfig, anthropic, prompt_template


@anthropic.call("claude-3-5-sonnet-20240620")
@prompt_template("Summarize this text: {text}")
def summarize(text: str): ...


@anthropic.call("claude-3-5-sonnet-20240620")
@prompt_template("Translate this text to {language}: {summary}")
def summarize_and_translate(text: str, language: str) -> BaseDynamicConfig:
    return {"computed_fields": {"summary": summarize(text)}}


response = summarize_and_translate("Long English text here...", "french")
print(response.content)
print(
    response.model_dump()["computed_fields"]
)  # This will contain the `summarize` response



from mirascope.core import Messages, mistral
from mistralai.client import MistralClient


@mistral.call("mistral-large-latest", client=MistralClient())
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



from anthropic import Anthropic
from mirascope.core import BaseMessageParam, anthropic


@anthropic.call("claude-3-5-sonnet-20240620", client=Anthropic())
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



from mirascope.core import anthropic
from tenacity import retry, stop_after_attempt, wait_exponential


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
@anthropic.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



from mirascope.core import BaseDynamicConfig, BaseMessageParam, gemini


@gemini.call("gemini-1.5-flash")
def recommend_book(genre: str) -> BaseDynamicConfig:
    return {
        "messages": [
            BaseMessageParam(role="user", content=f"Recommend a {genre} book")
        ],
        "call_params": {"max_tokens": 512},
        "metadata": {"tags": {"version:0001"}},
    }


response = recommend_book("fantasy")
print(response.content)



import asyncio
import base64
import io
from io import BytesIO
from typing import Any

from pydub import AudioSegment

SAMPLE_WIDTH = 2
FRAME_RATE = 24000
CHANNELS = 1


def audio_chunk_to_audio_segment(audio_bytes: bytes) -> AudioSegment:
    return AudioSegment.from_raw(
        io.BytesIO(audio_bytes),
        sample_width=SAMPLE_WIDTH,
        frame_rate=FRAME_RATE,
        channels=CHANNELS,
    )


def _encode_pcm_as_base64(audio_bytes: bytes | BytesIO) -> str:
    # Load the audio file from the byte stream
    audio = AudioSegment.from_file(
        io.BytesIO(audio_bytes) if isinstance(audio_bytes, bytes) else audio_bytes
    )

    # Resample to 24kHz mono pcm16
    pcm_audio = (
        audio.set_frame_rate(FRAME_RATE)
        .set_channels(CHANNELS)
        .set_sample_width(SAMPLE_WIDTH)
        .raw_data
    )

    # Encode to base64 string
    return base64.b64encode(pcm_audio).decode()


def audio_to_input_audio_buffer_append_event(
    audio_bytes: bytes | BytesIO,
) -> dict[str, Any]:
    return {
        "type": "input_audio_buffer.append",
        "audio": _encode_pcm_as_base64(audio_bytes),
    }


async def async_audio_input_audio_buffer_append_event(
    audio_bytes: bytes | BytesIO,
) -> dict[str, Any]:
    return await asyncio.to_thread(
        audio_to_input_audio_buffer_append_event, audio_bytes
    )


def audio_to_item_create_event(audio_bytes: bytes | BytesIO) -> dict[str, Any]:
    return {
        "type": "conversation.item.create",
        "item": {
            "type": "message",
            "role": "user",
            "content": [
                {"type": "input_audio", "audio": _encode_pcm_as_base64(audio_bytes)}
            ],
        },
    }


async def async_audio_to_item_create_event(
    audio_bytes: bytes | BytesIO,
) -> dict[str, Any]:
    return await asyncio.to_thread(audio_to_item_create_event, audio_bytes)



from mirascope.core import BaseMessageParam, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='user', content='Recommend a fantasy book')]



"""This module contains the type definition for the Cohere call keyword arguments."""

from cohere.types import Tool

from ..base import BaseCallKwargs
from .call_params import CohereCallParams


class CohereCallKwargs(CohereCallParams, BaseCallKwargs[Tool]):
    model: str
    message: str



"""The `extract_factory` method for generating provider specific create decorators."""

from collections.abc import Awaitable, Callable
from functools import wraps
from typing import ParamSpec, TypeVar, overload

from pydantic import BaseModel, ValidationError

from ._create import create_factory
from ._utils import (
    BaseType,
    GetJsonOutput,
    SameSyncAndAsyncClientSetupCall,
    SetupCall,
    extract_tool_return,
    fn_is_async,
    setup_extract_tool,
)
from ._utils._get_fields_from_call_args import get_fields_from_call_args
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_SameSyncAndAsyncClientT = TypeVar("_SameSyncAndAsyncClientT", contravariant=True)
_SyncBaseClientT = TypeVar("_SyncBaseClientT", contravariant=True)
_AsyncBaseClientT = TypeVar("_AsyncBaseClientT", contravariant=True)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_AsyncResponseT = TypeVar("_AsyncResponseT")
_AsyncResponseChunkT = TypeVar("_AsyncResponseChunkT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)
_P = ParamSpec("_P")


def extract_factory(  # noqa: ANN202
    *,
    TCallResponse: type[_BaseCallResponseT],
    TToolType: type[BaseTool],
    setup_call: SameSyncAndAsyncClientSetupCall[
        _SameSyncAndAsyncClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ]
    | SetupCall[
        _SyncBaseClientT,
        _AsyncBaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseT],
):
    """Returns the wrapped function with the provider specific interfaces."""
    create_decorator = create_factory(
        TCallResponse=TCallResponse, setup_call=setup_call
    )

    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _SyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _ResponseModelT | _ParsedOutputT]: ...

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _AsyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, Awaitable[_ResponseModelT | _ParsedOutputT]]: ...

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT]
        | Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _SyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        _ResponseModelT | _ParsedOutputT | Awaitable[_ResponseModelT | _ParsedOutputT],
    ]:
        fn._model = model  # pyright: ignore [reportFunctionMemberAccess]
        fn.__mirascope_call__ = True  # pyright: ignore [reportFunctionMemberAccess]
        tool = setup_extract_tool(response_model, TToolType)
        create_decorator_kwargs = {
            "model": model,
            "tools": [tool],
            "output_parser": None,
            "json_mode": json_mode,
            "client": client,
            "call_params": call_params,
        }

        if fn_is_async(fn):

            @wraps(fn)
            async def inner_async(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> _ResponseModelT:
                fields_from_call_args = get_fields_from_call_args(
                    response_model, fn, args, kwargs
                )
                call_response = await create_decorator(
                    fn=fn, **create_decorator_kwargs
                )(*args, **kwargs)
                json_output = get_json_output(call_response, json_mode)
                try:
                    output = extract_tool_return(
                        response_model, json_output, False, fields_from_call_args
                    )
                except ValidationError as e:
                    e._response = call_response  # pyright: ignore [reportAttributeAccessIssue]
                    raise e
                if isinstance(output, BaseModel):
                    output._response = call_response  # pyright: ignore [reportAttributeAccessIssue]
                return output if not output_parser else output_parser(output)  # pyright: ignore [reportArgumentType, reportReturnType]

            return inner_async
        else:

            @wraps(fn)
            def inner(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
                fields_from_call_args = get_fields_from_call_args(
                    response_model, fn, args, kwargs
                )
                call_response = create_decorator(fn=fn, **create_decorator_kwargs)(
                    *args, **kwargs
                )
                json_output = get_json_output(call_response, json_mode)
                try:
                    output = extract_tool_return(
                        response_model, json_output, False, fields_from_call_args
                    )
                except ValidationError as e:
                    e._response = call_response  # pyright: ignore [reportAttributeAccessIssue]
                    raise e
                if isinstance(output, BaseModel):
                    output._response = call_response  # pyright: ignore [reportAttributeAccessIssue]
                return output if not output_parser else output_parser(output)  # pyright: ignore [reportReturnType, reportArgumentType]

            return inner

    return decorator



"""usage docs: learn/calls.md#provider-specific-parameters"""

from __future__ import annotations

from typing import TYPE_CHECKING

from openai.types.chat import (
    ChatCompletionStreamOptionsParam,
    ChatCompletionToolChoiceOptionParam,
)
from openai.types.chat.completion_create_params import ResponseFormat
from typing_extensions import NotRequired

from ..base import BaseCallParams

if TYPE_CHECKING:
    from openai.types.chat.chat_completion_audio_param import (  # pyright: ignore [reportMissingImports]
        ChatCompletionAudioParam,
        ChatCompletionModality,  # pyright: ignore [reportAttributeAccessIssue]
    )
else:
    try:
        from openai.types.chat.chat_completion_audio_param import (  # pyright: ignore [reportMissingImports]
            ChatCompletionAudioParam,
            ChatCompletionModality,
        )
    except ImportError:

        class ChatCompletionAudioParam: ...

        class ChatCompletionModality: ...


class OpenAICallParams(BaseCallParams):
    """The parameters to use when calling the OpenAI API.

    [OpenAI API Reference](https://platform.openai.com/docs/api-reference/chat/create)

    Attributes:
        audio: ...
        frequency_penalty: ...
        logit_bias: ...
        logprobs: ...
        max_tokens: ...
        modalities: ...
        n: ...
        parallel_tool_calls: ...
        presence_penalty: ...
        response_format: ...
        seed: ...
        stop: ...
        stream_options: ...
        temperature: ...
        tool_choice: ...
        top_logprobs: ...
        top_p: ...
        user: ...
    """

    audio: NotRequired[ChatCompletionAudioParam | None]
    extra_headers: NotRequired[dict[str, str] | None]
    frequency_penalty: NotRequired[float | None]
    logit_bias: NotRequired[dict[str, int] | None]
    logprobs: NotRequired[bool | None]
    max_tokens: NotRequired[int | None]
    modalities: NotRequired[list[ChatCompletionModality] | None]
    n: NotRequired[int | None]
    parallel_tool_calls: NotRequired[bool]
    presence_penalty: NotRequired[float | None]
    response_format: NotRequired[ResponseFormat]
    seed: NotRequired[int | None]
    stop: NotRequired[str | list[str] | None]
    stream_options: NotRequired[ChatCompletionStreamOptionsParam | None]
    temperature: NotRequired[float | None]
    tool_choice: NotRequired[ChatCompletionToolChoiceOptionParam]
    top_logprobs: NotRequired[int | None]
    top_p: NotRequired[float | None]
    user: NotRequired[str]



from collections.abc import Awaitable, Callable, Sequence
from functools import wraps
from typing import ParamSpec, Protocol, TypeAlias, TypeVar, overload

from typing_extensions import TypeIs

from ..dynamic_config import BaseDynamicConfig
from ..message_param import BaseMessageParam
from ..messages import Messages
from ._convert_messages_to_message_params import (
    convert_messages_to_message_params,
)
from ._fn_is_async import fn_is_async

_P = ParamSpec("_P")

_MessageFuncReturnTypes: TypeAlias = BaseDynamicConfig | Messages.Type
_MessageFuncReturnT = TypeVar(
    "_MessageFuncReturnT", bound=_MessageFuncReturnTypes, contravariant=True
)

MessagesSyncFunction: TypeAlias = Callable[_P, _MessageFuncReturnT]
MessagesAsyncFunction: TypeAlias = Callable[_P, Awaitable[_MessageFuncReturnT]]


def _is_messages_type(value: object) -> TypeIs[Messages.Type]:
    return isinstance(
        value,
        str | Sequence | list | BaseMessageParam,
    )


class MessagesDecorator(Protocol):
    @overload
    def __call__(
        self,
        messages_fn: Callable[_P, Messages.Type],
    ) -> Callable[_P, list[BaseMessageParam]]: ...

    @overload
    def __call__(
        self,
        messages_fn: Callable[_P, BaseDynamicConfig],
    ) -> Callable[_P, BaseDynamicConfig]: ...

    @overload
    def __call__(
        self,
        messages_fn: Callable[_P, Awaitable[Messages.Type]],
    ) -> Callable[_P, Awaitable[list[BaseMessageParam]]]: ...

    @overload
    def __call__(
        self,
        messages_fn: Callable[_P, Awaitable[BaseDynamicConfig]],
    ) -> Callable[_P, Awaitable[BaseDynamicConfig]]: ...

    def __call__(
        self,
        messages_fn: MessagesSyncFunction[_P, _MessageFuncReturnT]
        | MessagesAsyncFunction[_P, _MessageFuncReturnT],
    ) -> (
        Callable[_P, list[BaseMessageParam] | BaseDynamicConfig]
        | Callable[_P, Awaitable[list[BaseMessageParam] | BaseDynamicConfig]]
    ): ...


def messages_decorator() -> MessagesDecorator:
    @overload
    def inner(
        messages_fn: Callable[_P, Messages.Type],
    ) -> Callable[_P, list[BaseMessageParam]]: ...

    @overload
    def inner(
        messages_fn: Callable[_P, BaseDynamicConfig],
    ) -> Callable[_P, BaseDynamicConfig]: ...

    @overload
    def inner(
        messages_fn: Callable[_P, Awaitable[Messages.Type]],
    ) -> Callable[_P, Awaitable[list[BaseMessageParam]]]: ...

    @overload
    def inner(
        messages_fn: Callable[_P, Awaitable[BaseDynamicConfig]],
    ) -> Callable[_P, Awaitable[BaseDynamicConfig]]: ...

    def inner(
        messages_fn: MessagesSyncFunction[_P, _MessageFuncReturnT]
        | MessagesAsyncFunction[_P, _MessageFuncReturnT],
    ) -> (
        Callable[_P, Awaitable[list[BaseMessageParam] | BaseDynamicConfig]]
        | Callable[_P, list[BaseMessageParam] | BaseDynamicConfig]
    ):
        if fn_is_async(messages_fn):

            @wraps(messages_fn)
            async def get_base_message_params_async(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> list[BaseMessageParam] | BaseDynamicConfig:
                raw_messages = await messages_fn(*args, **kwargs)
                if _is_messages_type(raw_messages):
                    return convert_messages_to_message_params(raw_messages)
                return raw_messages

            return get_base_message_params_async
        else:

            @wraps(messages_fn)
            def get_base_message_params(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> list[BaseMessageParam] | BaseDynamicConfig:
                raw_messages = messages_fn(*args, **kwargs)
                if _is_messages_type(raw_messages):
                    return convert_messages_to_message_params(raw_messages)
                return raw_messages

            return get_base_message_params

    return inner



from cohere import Client
from cohere.types import Tool, ToolParameterDefinitionsValue
from pydantic import BaseModel

client = Client()


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


def extract_book(text: str) -> Book:
    response = client.chat(
        model="command-r-plus",
        message=f"Extract {text}",
        tools=[
            Tool(
                name="Book",
                description="An extracted book.",
                parameter_definitions={
                    "title": ToolParameterDefinitionsValue(
                        description=None, type="string", required=True
                    ),
                    "author": ToolParameterDefinitionsValue(
                        description=None, type="string", required=True
                    ),
                },
            )
        ],
    )
    if response.tool_calls:
        return Book.model_validate(response.tool_calls[0].parameters)
    raise ValueError("No tool call found")


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



from mirascope.core import Messages, mistral
from mistralai import models


@mistral.call("mistral-large-latest")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


try:
    response = recommend_book("fantasy")
    print(response.content)
except models.HTTPValidationError as e:  # pyright: ignore [reportAttributeAccessIssue]
    # handle e.data: models.HTTPValidationErrorData
    raise (e)
except models.SDKError as e:  # pyright: ignore [reportAttributeAccessIssue]
    # handle exception
    raise (e)



from mirascope.core import prompt_template, vertex
from vertexai.generative_models import GenerativeModel


@vertex.call("", client=GenerativeModel(model_name="gemini-1.5-flash"))
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...



import inspect

from mirascope.core import Messages, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> Messages.Type:
    return inspect.cleandoc(
        f"""
        Recommend a {genre} book.
        Output in the format Title by Author.
        """
    )


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='system', content='Recommend a fantasy book.\nOutput in the format Title by Author.')]



from anthropic import Anthropic
from pydantic import BaseModel

client = Anthropic()


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


def extract_book(text: str) -> Book:
    message = client.messages.create(
        model="claude-3-5-sonnet-20240620",
        max_tokens=1024,
        messages=[{"role": "user", "content": f"Extract {text}"}],
        tools=[
            {
                "name": "Book",
                "description": "An extracted book.",
                "input_schema": {
                    "properties": {
                        "title": {"type": "string"},
                        "author": {"type": "string"},
                    },
                    "required": ["title", "author"],
                    "type": "object",
                },
            }
        ],
        tool_choice={"type": "tool", "name": "Book"},
    )
    for block in message.content:
        if block.type == "tool_use" and block.input is not None:
            return Book.model_validate(block.input)
    raise ValueError("No tool call found")


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



import asyncio

from mirascope.core import azure, prompt_template


@azure.call(model="gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
async def recommend_book(genre: str): ...


async def main():
    response = await recommend_book("fantasy")
    print(response.content)


asyncio.run(main())



from mirascope.core import openai
from openai import AsyncOpenAI


@openai.call("gpt-4o-mini", client=AsyncOpenAI())
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book"



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from PIL import Image
    from pydub import AudioSegment

    has_pil_module: bool
    has_pydub_module: bool
else:
    try:
        from PIL import Image  # pyright: ignore [reportAssignmentType]

        has_pil_module = True
    except ImportError:  # pragma: no cover
        has_pil_module = False

        class Image:
            class Image:
                def tobytes(self) -> bytes: ...

    try:
        from pydub import AudioSegment  # pyright: ignore [reportAssignmentType]

        has_pydub_module = True
    except ImportError:  # pragma: no cover
        has_pydub_module = False

        from io import FileIO

        class AudioSegment:
            def set_frame_rate(self, rate: int) -> AudioSegment: ...
            def set_channels(self, channels: int) -> AudioSegment: ...
            def set_sample_width(self, sample_width: int) -> AudioSegment: ...
            def export(self, format: str) -> FileIO: ...
            def read(self) -> bytes: ...



from mirascope.core import prompt_template


@prompt_template(
    """
    Recommend a book from one of the following genres:
    {genres:list}

    Examples:
    {examples:lists}
    """
)
def recommend_book_prompt(genres: list[str], examples: list[list[str]]): ...


prompt = recommend_book_prompt(
    genres=["fantasy", "scifi", "mystery"],
    examples=[
        ["Title: The Name of the Wind", "Author: Patrick Rothfuss"],
        ["Title: Mistborn: The Final Empire", "Author: Brandon Sanderson"],
    ],
)
print(prompt)
# Output: [
#     BaseMessageParam(
#         role="user",
#         content="Recommend a book from one of the following genres:\nfantasy\nscifi\nmystery\n\nExamples:\nTitle: The Name of the Wind\nAuthor: Patrick Rothfuss\n\nTitle: Mistborn: The Final Empire\nAuthor: Brandon Sanderson",
#     )
# ]

print(prompt[0].content)
# Output:
# Recommend a book from one of the following genres:
# fantasy
# scifi
# mystery

# Examples:
# Title: The Name of the Wind
# Author: Patrick Rothfuss

# Title: Mistborn: The Final Empire
# Author: Brandon Sanderson



from mirascope.core import BaseMessageParam, vertex
from vertexai.generative_models import GenerationConfig


@vertex.call(
    "gemini-1.5-flash",
    call_params={"generation_config": GenerationConfig(max_output_tokens=512)},
)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


response = recommend_book("fantasy")
print(response.content)



"""This module contains the setup_call function for Azure tools."""

import inspect
import os
import warnings
from collections.abc import AsyncIterable, Awaitable, Callable, Iterable
from typing import Any, cast, overload

from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.aio import ChatCompletionsClient as AsyncChatCompletionsClient
from azure.ai.inference.models import (
    ChatCompletions,
    ChatCompletionsResponseFormatJSON,
    ChatRequestMessage,
    StreamingChatCompletionsUpdate,
    UserMessage,
)
from azure.core.credentials import AzureKeyCredential

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import AsyncCreateFn, CreateFn, get_async_create_fn, get_create_fn
from ..call_kwargs import AzureCallKwargs
from ..call_params import AzureCallParams
from ..dynamic_config import AzureDynamicConfig
from ..tool import AzureTool, GenerateAzureStrictToolJsonSchema
from ._convert_message_params import convert_message_params
from ._get_credential import get_credential


@overload
def setup_call(
    *,
    model: str,
    client: AsyncChatCompletionsClient | None,
    fn: Callable[..., Awaitable[AzureDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AzureDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AzureCallParams,
    extract: bool,
) -> tuple[
    AsyncCreateFn[ChatCompletions, StreamingChatCompletionsUpdate],
    str | None,
    list[ChatRequestMessage],
    list[type[AzureTool]] | None,
    AzureCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: ChatCompletionsClient | None,
    fn: Callable[..., AzureDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: AzureDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AzureCallParams,
    extract: bool,
) -> tuple[
    CreateFn[ChatCompletions, StreamingChatCompletionsUpdate],
    str | None,
    list[ChatRequestMessage],
    list[type[AzureTool]] | None,
    AzureCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: ChatCompletionsClient | AsyncChatCompletionsClient | None,
    fn: Callable[..., AzureDynamicConfig]
    | Callable[..., Awaitable[AzureDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AzureDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AzureCallParams,
    extract: bool,
) -> tuple[
    CreateFn[ChatCompletions, StreamingChatCompletionsUpdate]
    | AsyncCreateFn[ChatCompletions, StreamingChatCompletionsUpdate],
    str | None,
    list[ChatRequestMessage],
    list[type[AzureTool]] | None,
    AzureCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, AzureTool, call_params
    )
    call_kwargs = cast(AzureCallKwargs, base_call_kwargs)
    messages = cast(list[BaseMessageParam | ChatRequestMessage], messages)
    messages = convert_message_params(messages)
    if json_mode:
        if tool_types and tool_types[0].model_config.get("strict", False):
            call_kwargs["response_format"] = ChatCompletionsResponseFormatJSON(
                {
                    "name": tool_types[0]._name(),
                    "description": tool_types[0]._description(),
                    "strict": True,
                    "schema": tool_types[0].model_json_schema(
                        schema_generator=GenerateAzureStrictToolJsonSchema
                    ),
                }
            )
        else:
            call_kwargs["response_format"] = ChatCompletionsResponseFormatJSON()
            json_mode_content = _utils.json_mode_content(
                tool_types[0] if tool_types else None
            ).strip()
            messages.append(UserMessage(content=json_mode_content))
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        if tool_types and tool_types[0].model_config.get("strict", False):
            warnings.warn(
                "You must set `json_mode=True` to use `strict=True` response models. "
                "Ignoring `strict` and using tools for extraction.",
                UserWarning,
            )
        call_kwargs["tool_choice"] = "required"
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        endpoint = os.environ["AZURE_INFERENCE_ENDPOINT"]
        credential = cast(AzureKeyCredential, get_credential())

        client = (
            AsyncChatCompletionsClient(endpoint=endpoint, credential=credential)
            if inspect.iscoroutinefunction(fn)
            else ChatCompletionsClient(endpoint=endpoint, credential=credential)
        )
    create = (
        get_async_create_fn(
            cast(
                AsyncCreateFn[
                    ChatCompletions, AsyncIterable[StreamingChatCompletionsUpdate]
                ],
                client.complete,
            )
        )
        if isinstance(client, AsyncChatCompletionsClient)
        else get_create_fn(
            cast(
                CreateFn[ChatCompletions, Iterable[StreamingChatCompletionsUpdate]],
                client.complete,
            )
        )
    )
    return create, prompt_template, messages, tool_types, call_kwargs



import json

from mirascope.core import Messages, groq


@groq.call("llama-3.1-70b-versatile", json_mode=True)
def get_book_info(book_title: str) -> Messages.Type:
    return Messages.User(f"Provide the author and genre of {book_title}")


response = get_book_info("The Name of the Wind")
print(json.loads(response.content))
# Output: {'author': 'Patrick Rothfuss', 'genre': 'Fantasy'}



from mirascope.core import Messages, cohere


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@cohere.call("command-r-plus", tools=[get_book_author])
def identify_author(
    book: str, history: list[cohere.CohereMessageParam]
) -> Messages.Type:
    messages = [*history]
    if book:
        messages.append(Messages.User(f"Who wrote {book}?"))
    return messages


history = []
response = identify_author("The Name of the Wind", history)
history += [response.user_message_param, response.message_param]
while tool := response.tool:
    tools_and_outputs = [(tool, tool.call())]
    history += response.tool_message_params(tools_and_outputs)
    response = identify_author("", history)
    history.append(response.message_param)
print(response.content)
# Output: The Name of the Wind was written by Patrick Rothfuss.



from mirascope.core import BaseMessageParam, bedrock
import boto3


@bedrock.call(
    "anthropic.claude-3-haiku-20240307-v1:0", client=boto3.client("bedrock-runtime")
)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



from mirascope.core import Messages, litellm
from pydantic import BaseModel


class Book(BaseModel):
    title: str
    author: str


@litellm.call("gpt-4o-mini", response_model=Book, stream=True)
def extract_book(text: str) -> Messages.Type:
    return Messages.User(f"Extract {text}")


book_stream = extract_book("The Name of the Wind by Patrick Rothfuss")
for partial_book in book_stream:
    print(partial_book)
# Output:
# title=None author=None
# title='' author=None
# title='The' author=None
# title='The Name' author=None
# title='The Name of' author=None
# title='The Name of the' author=None
# title='The Name of the Wind' author=None
# title='The Name of the Wind' author=''
# title='The Name of the Wind' author='Patrick'
# title='The Name of the Wind' author='Patrick Roth'
# title='The Name of the Wind' author='Patrick Rothf'
# title='The Name of the Wind' author='Patrick Rothfuss'



from mirascope.core import gemini
from mirascope.integrations.otel import with_hyperdx


@with_hyperdx()
@gemini.call("gemini-1.5-flash")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



from mirascope.core import BaseDynamicConfig, prompt_template


@prompt_template("Recommend a {uppercase_genre} book")
def recommend_book_prompt(genre: str) -> BaseDynamicConfig:
    uppercase_genre = genre.upper()
    return {
        "computed_fields": {"uppercase_genre": uppercase_genre},
    }


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='user', content='Recommend a FANTASY book')]



from mirascope.core import cohere


def parse_recommendation(response: cohere.CohereCallResponse) -> tuple[str, str]:
    title, author = response.content.split(" by ")
    return (title, author)


@cohere.call("command-r-plus", output_parser=parse_recommendation)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book. Output only Title by Author"


print(recommend_book("fantasy"))
# Output: ('"The Name of the Wind"', 'Patrick Rothfuss')



from mirascope.core import Messages, vertex
from vertexai.generative_models import GenerativeModel


@vertex.call("", client=GenerativeModel(model_name="gemini-1.5-flash"))
async def recommend_book_async(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



import inspect

from mirascope.core import BaseMessageParam, prompt_template


@prompt_template()
def recommend_book_prompt(
    genres: list[str], examples: list[list[str]]
) -> list[BaseMessageParam]:
    formatted_genres = "\n".join(genres)
    formatted_examples = "\n\n".join(["\n".join(e) for e in examples])
    return [
        BaseMessageParam(
            role="user",
            content=inspect.cleandoc(
                """
                Recommend a book from one of the following genres:
                {genres}

                Examples:
                {examples}
                """
            ).format(genres=formatted_genres, examples=formatted_examples),
        )
    ]


prompt = recommend_book_prompt(
    genres=["fantasy", "scifi", "mystery"],
    examples=[
        ["Title: The Name of the Wind", "Author: Patrick Rothfuss"],
        ["Title: Mistborn: The Final Empire", "Author: Brandon Sanderson"],
    ],
)
print(prompt)
# Output: [
#     BaseMessageParam(
#         role="user",
#         content="Recommend a book from one of the following genres:\nfantasy\nscifi\nmystery\n\nExamples:\nTitle: The Name of the Wind\nAuthor: Patrick Rothfuss\n\nTitle: Mistborn: The Final Empire\nAuthor: Brandon Sanderson",
#     )
# ]

print(prompt[0].content)
# Output:
# Recommend a book from one of the following genres:
# fantasy
# scifi
# mystery

# Examples:
# Title: The Name of the Wind
# Author: Patrick Rothfuss

# Title: Mistborn: The Final Empire
# Author: Brandon Sanderson



from mirascope.core import groq, prompt_template
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@groq.call("llama-3.1-70b-versatile")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



from groq import GroqError
from mirascope.core import BaseMessageParam, groq


@groq.call("llama-3.1-70b-versatile", stream=True)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except GroqError as e:
    print(f"Error: {str(e)}")



"""This module contains the type definition for the Vertex call keyword arguments."""

from vertexai.generative_models import Content, Tool

from ..base import BaseCallKwargs
from .call_params import VertexCallParams


class VertexCallKwargs(VertexCallParams, BaseCallKwargs[Tool]):
    contents: list[Content]



import io
import wave
from unittest.mock import Mock, patch

import pytest

from mirascope.core import BaseMessageParam
from mirascope.core.base import AudioPart, ImagePart, TextPart
from mirascope.core.base._utils._convert_messages_to_message_params import (
    _convert_message_sequence_part_to_content_part,
    _is_base_message_params,
    convert_message_content_to_message_param_content,
    convert_messages_to_message_params,
)


class MockAudioSegment:
    def set_frame_rate(self, frame_rate):
        return self

    def set_channels(self, channels):
        return self

    def set_sample_width(self, sample_width):
        return self

    def export(self, format):
        return io.BytesIO(b"mock_wav_data")


def test_convert_message_sequence_part_to_content_part_with_string():
    input_value = "Hello, World!"
    expected_output = TextPart(text="Hello, World!", type="text")
    result = _convert_message_sequence_part_to_content_part(input_value)
    assert result == expected_output


def test_convert_message_sequence_part_to_content_part_with_text_part():
    input_value = TextPart(type="text", text="Hello, World!")
    result = _convert_message_sequence_part_to_content_part(input_value)
    assert result == input_value


def test_convert_message_sequence_part_to_content_part_with_image_part():
    input_value = ImagePart(
        type="image",
        media_type="image/png",
        image=b"image_bytes",
        detail=None,
    )
    result = _convert_message_sequence_part_to_content_part(input_value)
    assert result == input_value


def test_convert_message_sequence_part_to_content_part_with_audio_part():
    input_value = AudioPart(
        type="audio",
        media_type="audio/wav",
        audio=b"audio_bytes",
    )
    result = _convert_message_sequence_part_to_content_part(input_value)
    assert result == input_value


def test_convert_message_sequence_part_to_content_part_with_pil_image():
    mock_image_instance = Mock()
    mock_image_instance.tobytes.return_value = b"image_bytes"

    from PIL import Image

    with (
        patch(
            "mirascope.core.base._utils._convert_messages_to_message_params.has_pil_module",
            True,
        ),
        patch(
            "mirascope.core.base._utils._convert_messages_to_message_params.Image.Image",
            Mock,
        ),
        patch(
            "mirascope.core.base._utils._convert_messages_to_message_params.get_image_type",
            return_value="png",
        ),
        patch(
            "mirascope.core.base._utils._convert_messages_to_message_params.isinstance",
            side_effect=lambda obj, cls: True
            if cls == Image.Image
            else isinstance(obj, cls),
        ),
    ):
        result = _convert_message_sequence_part_to_content_part(mock_image_instance)
        expected_output = ImagePart(
            type="image",
            media_type="image/png",
            image=b"image_bytes",
            detail=None,
        )
        assert result == expected_output


def test_convert_message_sequence_part_to_content_part_invalid_type():
    input_value = 12345
    with pytest.raises(ValueError) as exc_info:
        _convert_message_sequence_part_to_content_part(input_value)  # pyright: ignore [reportArgumentType]
    assert f"Invalid message sequence type: {input_value}" in str(exc_info.value)


def test_convert_message_sequence_to_content():
    input_sequence = [
        "Hello",
        TextPart(type="text", text="World"),
        ImagePart(
            type="image",
            media_type="image/jpeg",
            image=b"image_bytes",
            detail=None,
        ),
        AudioPart(
            type="audio",
            media_type="audio/mp3",
            audio=b"audio_bytes",
        ),
    ]
    expected_output = [
        TextPart(type="text", text="Hello"),
        TextPart(type="text", text="World"),
        ImagePart(
            type="image",
            media_type="image/jpeg",
            image=b"image_bytes",
            detail=None,
        ),
        AudioPart(
            type="audio",
            media_type="audio/mp3",
            audio=b"audio_bytes",
        ),
    ]
    result = convert_message_content_to_message_param_content(input_sequence)
    assert result == expected_output


def test_is_base_message_params_with_base_message_param_list():
    input_value = [
        BaseMessageParam(role="user", content="Hello"),
        BaseMessageParam(role="assistant", content="Hi there!"),
    ]
    result = _is_base_message_params(input_value)
    assert result is True


def test_is_base_message_params_with_invalid_list():
    input_value = ["Hello", "World"]
    result = _is_base_message_params(input_value)
    assert result is False


def test_convert_messages_to_message_params_with_string():
    messages = "Hello, World!"
    result = convert_messages_to_message_params(messages)
    expected_output = [BaseMessageParam(role="user", content="Hello, World!")]
    assert result == expected_output


def test_convert_messages_to_message_params_with_base_message_param():
    message_param = BaseMessageParam(role="user", content="Hello")
    result = convert_messages_to_message_params(message_param)
    assert result == [message_param]


def test_convert_messages_to_message_params_with_sequence():
    messages = ["Hello", "World"]

    expected_output = [
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Hello"),
                TextPart(type="text", text="World"),
            ],
        )
    ]
    result = convert_messages_to_message_params(messages)
    assert result == expected_output


def test_convert_messages_to_message_params_with_invalid_type():
    messages = 12345
    with pytest.raises(ValueError) as exc_info:
        convert_messages_to_message_params(messages)  # pyright: ignore [reportArgumentType]
    assert f"Invalid messages type: {messages}" in str(exc_info.value)


def test_convert_messages_to_message_params_with_custom_role():
    messages = "Hello, World!"
    result = convert_messages_to_message_params(messages, role="custom_role")
    expected_output = [BaseMessageParam(role="custom_role", content="Hello, World!")]
    assert result == expected_output


def test_convert_message_sequence_part_to_content_part_with_audio_segment():
    mock_audio_segment = MockAudioSegment()

    with (
        patch(
            "mirascope.core.base._utils._convert_messages_to_message_params.has_pydub_module",
            True,
        ),
        patch(
            "mirascope.core.base._utils._convert_messages_to_message_params.AudioSegment",
            MockAudioSegment,
        ),
    ):
        result = _convert_message_sequence_part_to_content_part(mock_audio_segment)  # pyright: ignore [reportArgumentType]

    expected_output = AudioPart(
        type="audio",
        media_type="audio/wav",
        audio=b"mock_wav_data",
    )
    assert result == expected_output


def test_convert_message_sequence_part_to_content_part_with_wave_read():
    wav_data = io.BytesIO()
    with wave.open(wav_data, "wb") as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(2)
        wav_file.setframerate(44100)
        wav_file.writeframes(b"\x00\x00" * 1000)  # 1000 samples of silence

    wav_data.seek(0)
    wave_read = wave.open(wav_data, "rb")

    result = _convert_message_sequence_part_to_content_part(wave_read)

    expected_output = AudioPart(
        type="audio",
        media_type="audio/wav",
        audio=b"\x00\x00" * 1000,
    )
    assert result == expected_output



from mirascope.core import BaseMessageParam, mistral
from mistralai.async_client import MistralAsyncClient


@mistral.call("mistral-large-latest", client=MistralAsyncClient())
async def recommend_book_async(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



from mirascope.core import prompt_template, vertex
from pydantic import BaseModel


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


@vertex.call("gemini-1.5-flash", response_model=Book, json_mode=True)
@prompt_template("Extract {text}")
def extract_book(text: str): ...


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



from anthropic import Anthropic
from mirascope.core import anthropic, prompt_template


@anthropic.call("claude-3-5-sonnet-20240620", client=Anthropic())
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...



from mirascope.core import BaseMessageParam, anthropic


@anthropic.call("claude-3-5-sonnet-20240620", response_model=list[str])
def extract_book(texts: list[str]) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Extract book titles from {texts}")]


book = extract_book(
    [
        "The Name of the Wind by Patrick Rothfuss",
        "Mistborn: The Final Empire by Brandon Sanderson",
    ]
)
print(book)
# Output: ["The Name of the Wind", "Mistborn: The Final Empire"]



from pydantic.fields import FieldInfo


class FromCallArgs:
    """Use this metadata to pass function arguments through to the response model.

    Example:

    ```python
    from typing import Annotated
    from pydantic import BaseModel, Field
    from mirascope.core import openai, prompt_template
    from mirascope.core.base import FromCallArgs

    class BookRecommendation(BaseModel):
        genre: Annotated[str, FromCallArgs()]
        title: str = Field(..., description="The title of the recommended book")
        author: str = Field(..., description="The author of the recommended book")

    @openai.call(model="gpt-4o-mini", response_model=BookRecommendation)
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book."
    ```
    """

    pass


def is_from_call_args(field: FieldInfo) -> bool:
    return any(isinstance(m, FromCallArgs) for m in field.metadata)



from mirascope.core import Messages, bedrock


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


response = recommend_book("fantasy")
print(response.content)



import asyncio

from mirascope.core import Messages, gemini


@gemini.call("gemini-1.5-flash", stream=True)
async def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


async def main():
    stream = await recommend_book("fantasy")
    async for chunk, _ in stream:
        print(chunk.content, end="", flush=True)


asyncio.run(main())



import io
import wave

from pydub.playback import play
from pydub import AudioSegment

from mirascope.core import openai


@openai.call(
    "gpt-4o-audio-preview",
    call_params={
        "audio": {"voice": "alloy", "format": "wav"},
        "modalities": ["text", "audio"],
    },
)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book(genre="fantasy")

print(response.audio_transcript)

if audio := response.audio:
    audio_io = io.BytesIO(audio)

    with wave.open(audio_io, "rb") as f:
        audio_segment = AudioSegment.from_raw(
            audio_io,
            sample_width=f.getsampwidth(),
            frame_rate=f.getframerate(),
            channels=f.getnchannels(),
        )

    play(audio_segment)



from proto.marshal.collections import RepeatedComposite
from pydantic import BaseModel
from vertexai.generative_models import (
    FunctionDeclaration,
    GenerativeModel,
    Tool,
    ToolConfig,
)

model = GenerativeModel("gemini-1.5-flash")


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


def extract_book(text: str) -> Book:
    response = model.generate_content(
        f"Extract {text}",
        tools=[
            Tool(
                function_declarations=[
                    FunctionDeclaration(
                        **{
                            "name": "Book",
                            "description": "An extracted book.",
                            "parameters": {
                                "properties": {
                                    "title": {"type": "string"},
                                    "author": {"type": "string"},
                                },
                                "required": ["title", "author"],
                                "type": "object",
                            },
                        }
                    )
                ]
            )
        ],
        tool_config=ToolConfig(
            function_calling_config=ToolConfig.FunctionCallingConfig(
                mode=ToolConfig.FunctionCallingConfig.Mode.ANY,
                allowed_function_names=["Book"],
            ),
        ),
    )
    if tool_calls := [
        part.function_call
        for candidate in response.candidates  # pyright: ignore [reportAttributeAccessIssue]
        for part in candidate.content.parts
        if part.function_call.args
    ]:
        return Book.model_validate(
            {
                k: v if not isinstance(v, RepeatedComposite) else list(v)
                for k, v in tool_calls[0].args.items()
            }
        )
    raise ValueError("No tool call found")


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



"""Tests the `_utils.convert_base_type_to_base_tool` module."""

from typing import Annotated

from mirascope.core.base._utils._convert_base_type_to_base_tool import (
    convert_base_type_to_base_tool,
)
from mirascope.core.base._utils._default_tool_docstring import DEFAULT_TOOL_DOCSTRING
from mirascope.core.base.tool import BaseTool


def test_convert_base_type_to_base_tool() -> None:
    """Tests the `convert_base_type_to_base_tool` function."""
    tool = convert_base_type_to_base_tool(int, BaseTool)
    assert tool._name() == "int"
    assert tool._description() == DEFAULT_TOOL_DOCSTRING
    assert "value" in tool.model_fields
    tool = convert_base_type_to_base_tool(Annotated[str, "a string"], BaseTool)  # type: ignore
    assert tool._name() == "str"
    assert tool._description() == DEFAULT_TOOL_DOCSTRING
    assert "value" in tool.model_fields



from typing import cast

from mirascope.core import cohere
from pydantic import BaseModel


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


@cohere.call("command-r-plus", response_model=Book)
def extract_book(text: str) -> str:
    return f"Extract {text}"


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'

response = cast(cohere.CohereCallResponse, book._response)  # pyright: ignore[reportAttributeAccessIssue]
print(response.model_dump())
# > {'metadata': {}, 'response': {'id': ...}, ...}



import boto3

bedrock_client = boto3.client(service_name="bedrock-runtime")


def recommend_book(genre: str) -> str:
    messages = [{"role": "user", "content": [{"text": f"Recommend a {genre} book"}]}]
    response = bedrock_client.converse(
        modelId="anthropic.claude-3-haiku-20240307-v1:0",
        messages=messages,
        inferenceConfig={"maxTokens": 1024},
    )
    output_message = response["output"]["message"]
    content = ""
    for content_piece in output_message["content"]:
        if "text" in content_piece:
            content += content_piece["text"]
    return content


output = recommend_book("fantasy")
print(output)



from mirascope.core import Messages, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> Messages.Type:
    return [
        Messages.System("You are a librarian"),
        Messages.User(f"Recommend a {genre} book"),
    ]


print(recommend_book_prompt("fantasy"))
# Output: [
#   BaseMessageParam(role='system', content='You are a librarian'),
#   BaseMessageParam(role='user', content='Recommend a fantasy book'),
# ]



from mirascope.core import Messages, cohere
from mirascope.integrations.otel import with_hyperdx


@with_hyperdx()
@cohere.call("command-r-plus")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


print(recommend_book("fantasy"))



"""Shared fixtures for the base module tests."""

from unittest.mock import AsyncMock, MagicMock

import pytest


@pytest.fixture()
def mock_setup_call() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = tuple([MagicMock() for _ in range(5)])
    return mock_setup_call


@pytest.fixture()
def mock_setup_call_async() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = tuple(
        [AsyncMock()] + [MagicMock() for _ in range(4)]
    )
    return mock_setup_call



from cohere import AsyncClient
from mirascope.core import cohere


@cohere.call("command-r-plus", client=AsyncClient())
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book"



from mirascope.core import prompt_template


@prompt_template("Recommend a {genre} book")
def recommend_book_prompt(genre: str): ...


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='user', content='Recommend a fantasy book')]



from mirascope.core import Messages, gemini, prompt_template
from pydantic import BaseModel


class Librarian(BaseModel):
    history: list[gemini.GeminiMessageParam] = []

    @gemini.call("gemini-1.5-flash")
    @prompt_template(
        """
        SYSTEM: You are a librarian
        MESSAGES: {self.history}
        USER: {query}
        """
    )
    def _call(self, query: str): ...

    def run(self) -> None:
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            print("(Assistant): ", end="", flush=True)
            response = self._call(query)
            print(response.content)
            self.history += [
                Messages.User(query),
                response.message_param,
            ]


Librarian().run()



"""Tests the `cohere._utils.handle_stream` module."""

import pytest
from cohere.types import (
    NonStreamedChatResponse,
    StreamedChatResponse,
    StreamEndStreamedChatResponse,
    TextGenerationStreamedChatResponse,
    ToolCall,
    ToolCallsGenerationStreamedChatResponse,
)

from mirascope.core.cohere._utils._handle_stream import (
    handle_stream,
    handle_stream_async,
)
from mirascope.core.cohere.tool import CohereTool


class FormatBook(CohereTool):
    """Returns the title and author nicely formatted."""

    title: str
    author: str

    def call(self) -> None:
        """Dummy call."""


@pytest.fixture()
def mock_chunks() -> list[StreamedChatResponse]:
    """Returns a list of mock `ChatCompletionChunk` instances."""
    tool_call = ToolCall(
        name="FormatBook",
        parameters={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
    )
    return [
        TextGenerationStreamedChatResponse(
            text="json_output",
        ),
        ToolCallsGenerationStreamedChatResponse(tool_calls=[tool_call]),
        StreamEndStreamedChatResponse(
            finish_reason="COMPLETE",
            response=NonStreamedChatResponse(generation_id="id", text="", meta=None),
        ),
    ]


def test_handle_stream(mock_chunks: list[StreamedChatResponse]) -> None:
    """Tests the `handle_stream` function."""

    result = list(handle_stream((c for c in mock_chunks), tool_types=[FormatBook]))
    assert len(result) == 3
    assert result[0][1] is None


@pytest.mark.asyncio
async def test_handle_stream_async(mock_chunks: list[StreamedChatResponse]) -> None:
    """Tests the `handle_stream_async` function."""

    async def generator():
        for chunk in mock_chunks:
            yield chunk

    result = []
    async for t in handle_stream_async(generator(), tool_types=[FormatBook]):
        result.append(t)
    assert len(result) == 3
    assert result[0][1] is None



import boto3
from pydantic import BaseModel

bedrock_client = boto3.client(service_name="bedrock-runtime")


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


def extract_book(text: str) -> Book:
    messages = [{"role": "user", "content": [{"text": f"Extract {text}"}]}]
    tool_config = {
        "tools": [
            {
                "toolSpec": {
                    "name": "Book",
                    "description": "An extracted book.",
                    "inputSchema": {
                        "json": {
                            "type": "object",
                            "properties": {
                                "title": {"type": "string"},
                                "author": {"type": "string"},
                            },
                            "required": ["title", "author"],
                        }
                    },
                }
            }
        ],
        "toolChoice": {"type": "tool", "name": "Book"},
    }
    response = bedrock_client.converse(
        modelId="anthropic.claude-3-haiku-20240307-v1:0",
        messages=messages,
        toolConfig=tool_config,
    )
    output_message = response["output"]["message"]
    messages.append(output_message)
    for content_piece in output_message["content"]:
        if "toolUse" in content_piece and content_piece["toolUse"].get("input"):
            tool_input = content_piece["toolUse"]["input"]
            return Book.model_validate(tool_input)
    raise ValueError("No tool call found")


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



from mirascope.core import BaseMessageParam, anthropic


def get_book_author(title: str) -> str:
    """Returns the author of the book with the given title

    Args:
        title: The title of the book.
    """
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@anthropic.call("claude-3-5-sonnet-20240620", tools=[get_book_author], stream=True)
def identify_authors(books: list[str]) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Who wrote {books}?")]


stream = identify_authors(["The Name of the Wind", "Mistborn: The Final Empire"])
for chunk, tool in stream:
    if tool:
        print(tool.call())
    else:
        print(chunk.content, end="", flush=True)



import asyncio

from mirascope.core import bedrock
from aiobotocore.session import get_session


async def get_async_client():
    session = get_session()
    async with session.create_client("bedrock-runtime") as client:
        return client


@bedrock.call(
    "anthropic.claude-3-haiku-20240307-v1:0", client=asyncio.run(get_async_client())
)
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book"



import json

from mistralai.client import MistralClient

client = MistralClient()


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


def identify_author(book: str) -> str:
    completion = client.chat(
        model="mistral-large-latest",
        messages=[{"role": "user", "content": f"Who wrote {book}?"}],
        tools=[
            {
                "function": {
                    "name": "get_book_author",
                    "description": "Returns the author of the book with the given title.",
                    "parameters": {
                        "properties": {"title": {"type": "string"}},
                        "required": ["title"],
                        "type": "object",
                    },
                },
                "type": "function",
            }
        ],
    )
    if tool_calls := completion.choices[0].message.tool_calls:
        if tool_calls[0].function.name == "get_book_author":
            return get_book_author(**json.loads(tool_calls[0].function.arguments))
    return completion.choices[0].message.content


author = identify_author("The Name of the Wind")
print(author)



from mirascope.core import BaseMessageParam, gemini
from mirascope.integrations.otel import with_hyperdx


@with_hyperdx()
@gemini.call("gemini-1.5-flash")
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


print(recommend_book("fantasy"))



"""Tests the `openai.call_response_chunk` module."""

import base64

from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import (
    Choice,
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)
from openai.types.completion_usage import CompletionUsage

from mirascope.core.openai.call_response_chunk import OpenAICallResponseChunk


def test_openai_call_response_chunk() -> None:
    """Tests the `OpenAICallResponseChunk` class."""
    tool_call = ChoiceDeltaToolCall(
        index=0,
        id="id",
        function=ChoiceDeltaToolCallFunction(
            arguments='{"key": "value"}', name="function"
        ),
        type="function",
    )
    choices = [
        Choice(
            delta=ChoiceDelta(content="content", tool_calls=[tool_call]),
            index=0,
            finish_reason="stop",
        )
    ]
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    chunk = ChatCompletionChunk(
        id="id",
        choices=choices,
        created=0,
        model="gpt-4o",
        object="chat.completion.chunk",
        usage=usage,
    )
    call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
    assert call_response_chunk.content == "content"
    assert call_response_chunk.finish_reasons == ["stop"]
    assert call_response_chunk.model == "gpt-4o"
    assert call_response_chunk.id == "id"
    assert call_response_chunk.usage == usage
    assert call_response_chunk.input_tokens == 1
    assert call_response_chunk.output_tokens == 1


def test_openai_call_response_chunk_no_choices_or_usage() -> None:
    """Tests the `OpenAICallResponseChunk` class with None values."""
    chunk = ChatCompletionChunk(
        id="id",
        choices=[],
        created=0,
        model="gpt-4o",
        object="chat.completion.chunk",
        usage=None,
    )
    call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
    assert call_response_chunk.content == ""
    assert call_response_chunk.usage is None
    assert call_response_chunk.input_tokens is None
    assert call_response_chunk.output_tokens is None


def test_openai_call_response_chunk_with_audio() -> None:
    """Tests the `OpenAICallResponseChunk` class with audio content."""
    audio_data = b"test audio data"
    audio_base64 = base64.b64encode(audio_data).decode()
    transcript = "This is a test transcript"

    choices = [
        Choice(
            delta=ChoiceDelta(
                content="content",
                audio={"data": audio_base64, "transcript": transcript},  # pyright: ignore [reportCallIssue]
            ),
            index=0,
            finish_reason="stop",
        )
    ]
    chunk = ChatCompletionChunk(
        id="id",
        choices=choices,
        created=0,
        model="gpt-4o",
        object="chat.completion.chunk",
    )
    call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
    assert call_response_chunk.audio == audio_data
    assert call_response_chunk.audio_transcript == transcript


def test_openai_call_response_chunk_with_partial_audio() -> None:
    """Tests the `OpenAICallResponseChunk` class with partial audio content."""
    choices = [
        Choice(
            delta=ChoiceDelta(
                content="content",
                audio={"transcript": "transcript only"},  # pyright: ignore [reportCallIssue]
            ),
            index=0,
            finish_reason="stop",
        )
    ]
    chunk = ChatCompletionChunk(
        id="id",
        choices=choices,
        created=0,
        model="gpt-4o",
        object="chat.completion.chunk",
    )
    call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
    assert call_response_chunk.audio is None
    assert call_response_chunk.audio_transcript == "transcript only"


def test_openai_call_response_chunk_without_audio() -> None:
    """Tests the `OpenAICallResponseChunk` class without audio content."""
    choices = [
        Choice(
            delta=ChoiceDelta(content="content"),  # no audio field
            index=0,
            finish_reason="stop",
        )
    ]
    chunk = ChatCompletionChunk(
        id="id",
        choices=choices,
        created=0,
        model="gpt-4o",
        object="chat.completion.chunk",
    )
    call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
    assert call_response_chunk.audio is None
    assert call_response_chunk.audio_transcript is None



import json

from mirascope.core import BaseDynamicConfig, Messages, cohere, prompt_template
from pydantic import BaseModel


class Book(BaseModel):
    title: str
    author: str


class Librarian(BaseModel):
    history: list[cohere.CohereMessageParam] = []
    library: list[Book] = [
        Book(title="The Name of the Wind", author="Patrick Rothfuss"),
        Book(title="Mistborn: The Final Empire", author="Brandon Sanderson"),
    ]

    def _available_books(self) -> str:
        """Returns the list of books available in the library."""
        return json.dumps([book.model_dump() for book in self.library])

    @cohere.call("command-r-plus", stream=True)
    @prompt_template(
        """
        SYSTEM: You are a librarian
        MESSAGES: {self.history}
        USER: {query}
        """
    )
    def _stream(self, query: str) -> BaseDynamicConfig:
        return {"tools": [self._available_books]}

    def _step(self, query: str) -> None:
        if query:
            self.history.append(Messages.User(query))
        stream = self._stream(query)
        tools_and_outputs = []
        for chunk, tool in stream:
            if tool:
                print(f"[Calling Tool '{tool._name()}' with args {tool.args}]")
                tools_and_outputs.append((tool, tool.call()))
            else:
                print(chunk.content, end="", flush=True)
        self.history.append(stream.message_param)
        if tools_and_outputs:
            self.history += stream.tool_message_params(tools_and_outputs)
            self._step("")

    def run(self) -> None:
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            print("(Assistant): ", end="", flush=True)
            self._step(query)
            print()


Librarian().run()



from mirascope.core import openai, prompt_template
from openai import OpenAI


@openai.call("gpt-4o-mini", client=OpenAI())
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...



from mirascope.core import anthropic


def get_book_author(title: str) -> str:
    """Returns the author of the book with the given title

    Args:
        title: The title of the book.
    """
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@anthropic.call("claude-3-5-sonnet-20240620", tools=[get_book_author])
def identify_author(book: str) -> str:
    return f"Who wrote {book}?"


response = identify_author("The Name of the Wind")
if tool := response.tool:
    print(tool.call())
    # Output: Patrick Rothfuss
    print(f"Original tool call: {tool.tool_call}")
else:
    print(response.content)



import builtins
from typing import Any


def define_env(env: Any) -> None:
    @env.filter
    def provider_dir(provider: Any) -> str:
        # Convert to lowercase
        lower_provider = provider.lower()
        # Split by spaces or hyphens and take the first part
        first_word = lower_provider.split()[0]
        return first_word

    @env.macro
    def upper(v: str) -> str:
        return v.upper()

    @env.macro
    def zip(*args: list) -> Any:
        return builtins.zip(*args, strict=True)



"""Tests the `vertex._utils.get_json_output` module."""

import pytest
from google.cloud.aiplatform_v1beta1.types import FunctionCall
from vertexai.generative_models import (
    Candidate,
    Content,
    Part,
)
from vertexai.generative_models import (
    GenerationResponse as GenerateContentResponse,
)

from mirascope.core.vertex._utils._get_json_output import get_json_output
from mirascope.core.vertex.call_response import VertexCallResponse
from mirascope.core.vertex.call_response_chunk import VertexCallResponseChunk


@pytest.fixture()
def mock_generate_content_response() -> GenerateContentResponse:
    """Returns a mock `GenerateContentResponse` instance."""
    raw_part = Part()
    raw_part._raw_part.function_call = FunctionCall(
        name="FormatBook",
        args={
            "title": "The Name of the Wind",
            "author": "Patrick Rothfuss",
        },
    )
    return GenerateContentResponse.from_dict(
        {
            "candidates": [
                Candidate.from_dict(
                    {
                        "finish_reason": 1,
                        "content": Content(
                            parts=[
                                Part.from_text("json_output"),
                                raw_part,
                            ],
                            role="model",
                        ).to_dict(),
                    }
                ).to_dict()
            ]
        }
    )


def test_get_json_output_call_response(
    mock_generate_content_response: GenerateContentResponse,
) -> None:
    """Tests the `get_json_output` function with a call response."""
    call_response = VertexCallResponse(
        metadata={},
        response=mock_generate_content_response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert get_json_output(call_response, json_mode=True) == "json_output"
    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    mock_generate_content_response._raw_response.candidates[0].content.parts.pop(1)
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk(
    mock_generate_content_response: GenerateContentResponse,
) -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    call_response_chunk = VertexCallResponseChunk(chunk=mock_generate_content_response)
    assert get_json_output(call_response_chunk, json_mode=True) == "json_output"

    with pytest.raises(
        ValueError, match="Vertex only supports structured streaming in json mode."
    ):
        get_json_output(call_response_chunk, json_mode=False)



"""Tests the `cohere.stream` module."""

import pytest
from cohere.types import (
    ApiMeta,
    ApiMetaBilledUnits,
    ChatMessage,
    NonStreamedChatResponse,
    StreamEndStreamedChatResponse,
    StreamStartStreamedChatResponse,
    TextGenerationStreamedChatResponse,
)

from mirascope.core.cohere.call_response import CohereCallResponse
from mirascope.core.cohere.call_response_chunk import CohereCallResponseChunk
from mirascope.core.cohere.stream import CohereStream


def test_cohere_stream() -> None:
    """Tests the `CohereStream` class."""
    assert CohereStream._provider == "cohere"
    chunks = [
        StreamStartStreamedChatResponse(generation_id="id"),
        TextGenerationStreamedChatResponse(
            text="content",
        ),
        StreamEndStreamedChatResponse(
            finish_reason="COMPLETE",
            response=NonStreamedChatResponse(
                generation_id="id",
                text="content",
                meta=ApiMeta(
                    billed_units=ApiMetaBilledUnits(input_tokens=1, output_tokens=1)
                ),
            ),
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = CohereCallResponseChunk(chunk=chunk)
            yield call_response_chunk, None

    stream = CohereStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=CohereCallResponse,
        model="command-r-plus",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[ChatMessage(role="CHATBOT", message="content")],  # type: ignore
        call_params={},
        call_kwargs={},
    )

    with pytest.raises(
        ValueError, match="No stream response, check if the stream has been consumed."
    ):
        stream.construct_call_response()

    assert stream.cost is None
    for _ in stream:
        pass
    assert stream.cost == 1.8e-5
    assert stream.message_param == ChatMessage(
        role="assistant",  # type: ignore
        message="content",
        tool_calls=None,
    )


def test_construct_call_response() -> None:
    chunks = [
        StreamStartStreamedChatResponse(generation_id="id"),
        TextGenerationStreamedChatResponse(
            text="content",
        ),
        StreamEndStreamedChatResponse(
            finish_reason="COMPLETE",
            response=NonStreamedChatResponse(
                generation_id="id",
                text="content",
                meta=ApiMeta(
                    billed_units=ApiMetaBilledUnits(input_tokens=1, output_tokens=1)
                ),
            ),
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = CohereCallResponseChunk(chunk=chunk)
            yield call_response_chunk, None

    stream = CohereStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=CohereCallResponse,
        model="command-r-plus",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[ChatMessage(role="CHATBOT", message="content")],  # type: ignore
        call_params={},
        call_kwargs={},
    )
    assert stream.cost is None
    for _ in stream:
        pass
    constructed_call_response = stream.construct_call_response()
    usage = ApiMetaBilledUnits(input_tokens=1, output_tokens=1)
    completion = NonStreamedChatResponse(
        generation_id="id",
        text="content",
        finish_reason="COMPLETE",
        meta=ApiMeta(billed_units=usage),
    )
    call_response = CohereCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert constructed_call_response.response == call_response.response


def test_construct_call_response_no_usage() -> None:
    """Tests the `GroqStream.construct_call_response` method with no usage."""
    chunks = [
        StreamStartStreamedChatResponse(generation_id="id"),
        TextGenerationStreamedChatResponse(
            text="content",
        ),
        StreamEndStreamedChatResponse(
            finish_reason="COMPLETE",
            response=NonStreamedChatResponse(
                generation_id="id", text="content", meta=None
            ),
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = CohereCallResponseChunk(chunk=chunk)
            yield call_response_chunk, None

    stream = CohereStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=CohereCallResponse,
        model="command-r-plus",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[ChatMessage(role="CHATBOT", message="content")],  # type: ignore
        call_params={},
        call_kwargs={},
    )
    for _ in stream:
        pass
    constructed_call_response = stream.construct_call_response()
    assert constructed_call_response.usage is None



from cohere import AsyncClient
from mirascope.core import cohere, prompt_template


@cohere.call("command-r-plus", client=AsyncClient())
@prompt_template("Recommend a {genre} book")
async def recommend_book_async(genre: str): ...



from mirascope.core import BaseMessageParam, prompt_template


@prompt_template()
def chatbot(query: str, history: list[BaseMessageParam]) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(role="system", content="You are a librarian"),
        *history,
        BaseMessageParam(role="user", content=query),
    ]


history = [
    BaseMessageParam(role="user", content="Recommend a book"),
    BaseMessageParam(role="assistant", content="What genre do you like?"),
]
print(chatbot("fantasy", history))
# Output: [
#     BaseMessageParam(role="system", content="You are a librarian"),
#     BaseMessageParam(role="user", content="Recommend a book"),
#     BaseMessageParam(role="assistant", content="What genre do you like?"),
#     BaseMessageParam(role="user", content="fantasy"),
# ]



"""This module defines the function return type for functions as LLM calls."""

from vertexai.generative_models import Content

from ..base import BaseDynamicConfig, BaseMessageParam
from .call_params import VertexCallParams

VertexDynamicConfig = BaseDynamicConfig[Content | BaseMessageParam, VertexCallParams]
"""The function return type for functions wrapped with the `vertex_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.vertex import VertexDynamicConfig, vertex_call


@vertex_call("gemini-1.5-flash")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> VertexDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



from mirascope.core import Messages, mistral
from pydantic import BaseModel


class Librarian(BaseModel):
    history: list[mistral.MistralMessageParam] = []

    @mistral.call("mistral-large-latest")
    def _call(self, query: str) -> Messages.Type:
        return [
            Messages.System("You are a librarian"),
            *self.history,
            Messages.User(query),
        ]

    def run(self) -> None:
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            print("(Assistant): ", end="", flush=True)
            response = self._call(query)
            print(response.content)
            self.history += [
                Messages.User(query),
                response.message_param,
            ]


Librarian().run()



from mirascope.core import openai, prompt_template
from openai import AsyncOpenAI


@openai.call("gpt-4o-mini", client=AsyncOpenAI())
@prompt_template("Recommend a {genre} book")
async def recommend_book_async(genre: str): ...



from groq import Groq

client = Groq()


def recommend_book(genre: str) -> str:
    completion = client.chat.completions.create(
        model="llama-3.1-70b-versatile",
        messages=[{"role": "user", "content": f"Recommend a {genre} book"}],
    )
    return str(completion.choices[0].message.content)


output = recommend_book("fantasy")
print(output)



import json

from mirascope.core import gemini, prompt_template


@gemini.call("gemini-1.5-flash", json_mode=True)
@prompt_template("Provide the author and genre of {book_title}")
def get_book_info(book_title: str): ...


response = get_book_info("The Name of the Wind")
print(json.loads(response.content))
# Output: {'author': 'Patrick Rothfuss', 'genre': 'Fantasy'}



from mirascope.core import BaseTool, cohere
from pydantic import Field


class GetBookAuthor(BaseTool):
    """Returns the author of the book with the given title."""

    title: str = Field(..., description="The title of the book.")

    def call(self) -> str:
        if self.title == "The Name of the Wind":
            return "Patrick Rothfuss"
        elif self.title == "Mistborn: The Final Empire":
            return "Brandon Sanderson"
        else:
            return "Unknown"


@cohere.call("command-r-plus", tools=[GetBookAuthor], stream=True)
def identify_authors(books: list[str]) -> str:
    return f"Who wrote {books}?"


stream = identify_authors(["The Name of the Wind", "Mistborn: The Final Empire"])
for chunk, tool in stream:
    if tool:  # will always be None, not supported at the provider level
        print(tool.call())
    else:
        print(chunk.content, end="", flush=True)



from mirascope.core import prompt_template


@prompt_template("Here's an audio book snippet: {audio_wave:audio} What book is this?")
def identify_book_prompt(audio_wave: bytes): ...


print(identify_book_prompt(b"..."))
# Output: [
#     BaseMessageParam(
#         role="user",
#         content=[
#             TextPart(type="text", text="Here's an audio book snippet:"),
#             AudioPart(type='audio', media_type='audio/wav', audio=b'...'),
#             TextPart(type="text", text="What book is this?"),
#         ],
#     )
# ]



import json

from mirascope.core import BaseDynamicConfig, Messages, litellm
from pydantic import BaseModel


class Book(BaseModel):
    title: str
    author: str


class Librarian(BaseModel):
    history: list[litellm.OpenAIMessageParam] = []
    library: list[Book] = [
        Book(title="The Name of the Wind", author="Patrick Rothfuss"),
        Book(title="Mistborn: The Final Empire", author="Brandon Sanderson"),
    ]

    def _available_books(self) -> str:
        """Returns the list of books available in the library."""
        return json.dumps([book.model_dump() for book in self.library])

    @litellm.call("gpt-4o-mini")
    def _call(self, query: str) -> BaseDynamicConfig:
        messages = [
            Messages.System("You are a librarian"),
            *self.history,
            Messages.User(query),
        ]
        return {"messages": messages, "tools": [self._available_books]}

    def _step(self, query: str) -> str:
        if query:
            self.history.append(Messages.User(query))
        response = self._call(query)
        self.history.append(response.message_param)
        tools_and_outputs = []
        if tools := response.tools:
            for tool in tools:
                print(f"[Calling Tool '{tool._name()}' with args {tool.args}]")
                tools_and_outputs.append((tool, tool.call()))
            self.history += response.tool_message_params(tools_and_outputs)
            return self._step("")
        else:
            return response.content

    def run(self) -> None:
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            print("(Assistant): ", end="", flush=True)
            step_output = self._step(query)
            print(step_output)


Librarian().run()



from mirascope.core import bedrock, prompt_template
import boto3


@bedrock.call(
    "anthropic.claude-3-haiku-20240307-v1:0", client=boto3.client("bedrock-runtime")
)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...



def calculate_recall_precision(output: str, expected: str) -> tuple[float, float]:
    output_words = set(output.lower().split())
    expected_words = set(expected.lower().split())

    common_words = output_words.intersection(expected_words)

    recall = len(common_words) / len(expected_words) if expected_words else 0
    precision = len(common_words) / len(output_words) if output_words else 0

    return recall, precision


# Example usage
output = "The Eiffel Tower is a famous landmark in Paris, France."
expected = (
    "The Eiffel Tower, located in Paris, is an iron lattice tower on the Champ de Mars."
)
recall, precision = calculate_recall_precision(output, expected)
print(f"Recall: {recall:.2f}, Precision: {precision:.2f}")
# Output: Recall: 0.40, Precision: 0.60



import re


def contains_email(output: str) -> bool:
    email_pattern = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"
    return bool(re.search(email_pattern, output))


# Example usage
output = "My email is john.doe@example.com"
print(contains_email(output))
# Output: True



"""Tests the `cohere.call_response_chunk` module."""

from cohere.types import (
    ApiMeta,
    ApiMetaBilledUnits,
    NonStreamedChatResponse,
    StreamEndStreamedChatResponse,
    StreamStartStreamedChatResponse,
    TextGenerationStreamedChatResponse,
)

from mirascope.core.cohere.call_response_chunk import CohereCallResponseChunk


def test_cohere_call_response_chunk() -> None:
    """Tests the `CohereCallResponseChunk` class."""
    usage = ApiMetaBilledUnits(input_tokens=1, output_tokens=1)
    chunk_start = StreamStartStreamedChatResponse(generation_id="id")
    chunk = TextGenerationStreamedChatResponse(
        text="content",
    )
    chunk_finish = StreamEndStreamedChatResponse(
        finish_reason="COMPLETE",
        response=NonStreamedChatResponse(
            generation_id="id",
            text="content",
            meta=ApiMeta(billed_units=usage),
        ),
    )
    call_response_chunk_start = CohereCallResponseChunk(chunk=chunk_start)
    call_response_chunk = CohereCallResponseChunk(chunk=chunk)
    call_response_chunk_finish = CohereCallResponseChunk(chunk=chunk_finish)
    assert call_response_chunk_start.id == "id"
    assert call_response_chunk.content == "content"
    assert call_response_chunk.finish_reasons is None
    assert call_response_chunk.model is None
    assert call_response_chunk.id is None
    assert call_response_chunk.usage is None
    assert call_response_chunk.input_tokens is None
    assert call_response_chunk.output_tokens is None
    assert call_response_chunk_finish.content == ""
    assert call_response_chunk_finish.id == "id"
    assert call_response_chunk_finish.usage == usage
    assert call_response_chunk_finish.input_tokens == 1
    assert call_response_chunk_finish.output_tokens == 1
    assert call_response_chunk_finish.finish_reasons == ["COMPLETE"]



"""Utility for converting `BaseMessageParam` to `BedrockMessageParam`."""

from typing import cast

from ...base import BaseMessageParam
from .._types import ConversationRoleType, InternalBedrockMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | InternalBedrockMessageParam],
) -> list[InternalBedrockMessageParam]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif isinstance((content := message_param.content), str):
            converted_message_params.append(
                {
                    "role": cast(ConversationRoleType, message_param.role),
                    "content": [{"text": content}],
                }
            )
        else:
            converted_content = []
            for part in content:
                if part.type == "text":
                    converted_content.append({"text": part.text})
                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/gif",
                        "image/webp",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. Bedrock"
                            " currently only supports JPEG, PNG, GIF, and WebP images."
                        )
                    converted_content.append(
                        {"format": part.media_type, "bytes": part.image}
                    )
                else:
                    raise ValueError(
                        "Bedrock currently only supports text and image parts. "
                        f"Part provided: {part.type}"
                    )
            converted_message_params.append(
                {"role": message_param.role, "content": converted_content}
            )
    return converted_message_params



from mirascope.core import vertex


@vertex.call(model="gemini-1.5-flash", stream=True)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except Exception as e:
    print(f"Error: {str(e)}")



"""Tests the `azure._utils.calculate_cost` function."""

from mirascope.core.azure._utils._calculate_cost import calculate_cost


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function."""
    assert calculate_cost(None, None, model="gpt-4o-mini") is None
    assert calculate_cost(1, 1, model="unknown") is None
    assert calculate_cost(1, 1, model="gpt-4o-mini") is None



from mirascope.core import Messages, prompt_template
from PIL import Image


@prompt_template()
def recommend_book_prompt(previous_book: Image.Image) -> Messages.Type:
    return Messages.User(
        ["I just read this book:", previous_book, "What should I read next?"]
    )


with Image.open("...") as image:
    print(recommend_book_prompt(image))
# Output: [
#     BaseMessageParam(
#         role="user",
#         content=[
#             TextPart(type="text", text="I just read this book:"),
#             ImagePart(type="image", media_type="image/jpeg", image=b"...", detail=None),
#             TextPart(type="text", text="What should I read next?"),
#         ],
#     )
# ]



from mirascope.core import BaseMessageParam
from mirascope.core.base import ImagePart, TextPart


def recommend_book_prompt(previous_book_jpeg: bytes) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="I just read this book:"),
                ImagePart(
                    type="image",
                    media_type="image/jpeg",
                    image=previous_book_jpeg,
                    detail=None,
                ),
                TextPart(type="text", text="What should I read next?"),
            ],
        )
    ]


print(recommend_book_prompt(b"..."))
# Output: [
#     BaseMessageParam(
#         role="user",
#         content=[
#             TextPart(type="text", text="I just read this book:"),
#             ImagePart(type="image", media_type="image/jpeg", image=b"...", detail=None),
#             TextPart(type="text", text="What should I read next?"),
#         ],
#     )
# ]



from mirascope.core import BaseDynamicConfig, gemini, prompt_template


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@gemini.call("gemini-1.5-flash", tools=[get_book_author])
@prompt_template(
    """
    MESSAGES: {history}
    USER: {query}
    """
)
def identify_author(
    book: str, history: list[gemini.GeminiMessageParam]
) -> BaseDynamicConfig:
    return {"computed_fields": {"query": f"Who wrote {book}" if book else ""}}


history = []
response = identify_author("The Name of the Wind", history)
history += [response.user_message_param, response.message_param]
while tool := response.tool:
    tools_and_outputs = [(tool, tool.call())]
    history += response.tool_message_params(tools_and_outputs)
    response = identify_author("", history)
    history.append(response.message_param)
print(response.content)
# Output: The Name of the Wind was written by Patrick Rothfuss.



from mirascope.core import Messages, anthropic


@anthropic.call("claude-3-5-sonnet-20240620", call_params={"max_tokens": 512})
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


response = recommend_book("fantasy")
print(response.content)



"""Tests the `bedrock.call_response_chunk` module."""

from mypy_boto3_bedrock_runtime.type_defs import TokenUsageTypeDef

from mirascope.core.bedrock._types import StreamOutputChunk
from mirascope.core.bedrock.call_response_chunk import BedrockCallResponseChunk


def test_bedrock_call_response_chunk() -> None:
    """Tests the `BedrockCallResponseChunk` class."""
    usage = TokenUsageTypeDef(inputTokens=1, outputTokens=1)  # pyright: ignore [reportCallIssue]
    chunks = [
        StreamOutputChunk(
            messageStart={},  # pyright: ignore [reportArgumentType]
            responseMetadata={"RequestId": "id"},  # pyright: ignore [reportArgumentType]
            model="anthropic.claude-3-haiku-20240307-v1:0",
        ),
        StreamOutputChunk(
            contentBlockDelta={
                "contentBlockIndex": 0,
                "delta": {"text": "content"},
            },
            responseMetadata={"RequestId": "id"},  # pyright: ignore [reportArgumentType]
            model="anthropic.claude-3-haiku-20240307-v1:0",
        ),
        StreamOutputChunk(
            messageStop={"stopReason": "end_turn"},
            responseMetadata={"RequestId": "id"},  # pyright: ignore [reportArgumentType]
            model="anthropic.claude-3-haiku-20240307-v1:0",
            metadata={"usage": usage},  # pyright: ignore [reportArgumentType]
        ),
    ]

    call_response_chunk_0 = BedrockCallResponseChunk(chunk=chunks[0])
    call_response_chunk_1 = BedrockCallResponseChunk(chunk=chunks[1])
    call_response_chunk_2 = BedrockCallResponseChunk(chunk=chunks[2])

    # Test chunk 0 (message start)
    assert call_response_chunk_0.content == ""
    assert call_response_chunk_0.finish_reasons == []
    assert call_response_chunk_0.model == "anthropic.claude-3-haiku-20240307-v1:0"
    assert call_response_chunk_0.id == "id"
    assert call_response_chunk_0.usage is None
    assert call_response_chunk_0.input_tokens is None
    assert call_response_chunk_0.output_tokens is None

    # Test chunk 1 (content delta)
    assert call_response_chunk_1.content == "content"
    assert call_response_chunk_1.finish_reasons == []
    assert call_response_chunk_1.model == "anthropic.claude-3-haiku-20240307-v1:0"
    assert call_response_chunk_1.id == "id"
    assert call_response_chunk_1.usage is None
    assert call_response_chunk_1.input_tokens is None
    assert call_response_chunk_1.output_tokens is None

    # Test chunk 2 (message stop)
    assert call_response_chunk_2.content == ""
    assert call_response_chunk_2.finish_reasons == ["end_turn"]
    assert call_response_chunk_2.model == "anthropic.claude-3-haiku-20240307-v1:0"
    assert call_response_chunk_2.id == "id"
    assert call_response_chunk_2.usage == usage
    assert call_response_chunk_2.input_tokens == 1
    assert call_response_chunk_2.output_tokens == 1



from mirascope.core import mistral, prompt_template


@mistral.call("mistral-large-latest")
@prompt_template("Summarize this text: {text}")
def summarize(text: str): ...


@mistral.call("mistral-large-latest")
@prompt_template("Translate this text to {language}: {text}")
def translate(text: str, language: str): ...


summary = summarize("Long English text here...")
translation = translate(summary.content, "french")
print(translation.content)



"""The `BedrockStream` class for convenience around streaming LLM calls.

usage docs: learn/streams.md
"""

from typing import Any, cast

from mypy_boto3_bedrock_runtime.literals import StopReasonType as FinishReason
from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef,
    ConverseStreamMetadataEventTypeDef,
    MessageOutputTypeDef,
    ResponseMetadataTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseStreamMetadataEventTypeDef as AsyncConverseStreamMetadataEventTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ResponseMetadataTypeDef as AsyncResponseMetadataTypeDef,
)

from ..base.stream import BaseStream
from ._types import (
    AssistantMessageTypeDef,
    InternalBedrockMessageParam,
    ToolTypeDef,
    ToolUseBlockContentTypeDef,
    ToolUseBlockMessageTypeDef,
    UserMessageTypeDef,
)
from ._utils import (
    calculate_cost,
)
from .call_params import BedrockCallParams
from .call_response import BedrockCallResponse
from .call_response_chunk import BedrockCallResponseChunk
from .dynamic_config import BedrockDynamicConfig
from .tool import BedrockTool

_DEFAULT_RESPONSE_METADATA = ResponseMetadataTypeDef(
    RequestId="", HTTPStatusCode=500, HTTPHeaders={}, RetryAttempts=0
)


class BedrockStream(
    BaseStream[
        BedrockCallResponse,
        BedrockCallResponseChunk,
        UserMessageTypeDef,
        AssistantMessageTypeDef,
        ToolUseBlockMessageTypeDef,
        InternalBedrockMessageParam,
        BedrockTool,
        ToolTypeDef,
        BedrockDynamicConfig,
        BedrockCallParams,
        FinishReason,
    ]
):
    """A class for convenience around streaming Bedrock LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.bedrock import bedrock_call


    @bedrock_call("gpt-4o-mini", stream=True)
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    stream = recommend_book("fantasy")  # returns `BedrockStream` instance
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    _provider = "bedrock"

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        super().__init__(*args, **kwargs)
        self._metadata: (
            ConverseStreamMetadataEventTypeDef
            | AsyncConverseStreamMetadataEventTypeDef
            | None
        ) = None
        self._response_metadata: (
            ResponseMetadataTypeDef | AsyncResponseMetadataTypeDef
        ) = _DEFAULT_RESPONSE_METADATA

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self,
        tool_calls: list[ToolUseBlockContentTypeDef] | None = None,
        content: str | None = None,
    ) -> AssistantMessageTypeDef:
        """Constructs the message parameter for the assistant."""
        contents = []
        if content is not None:
            contents.append({"text": content})

        if tool_calls:
            contents += [
                cast(ToolUseBlockContentTypeDef, {"toolUse": tool_call})
                for tool_call in tool_calls
            ]
        message_param = AssistantMessageTypeDef(role="assistant", content=contents)
        return message_param

    def _update_properties(self, chunk: BedrockCallResponseChunk) -> None:
        """Updates the properties of the stream."""
        super()._update_properties(chunk)
        if metadata := chunk.chunk.get("metadata"):
            self._metadata = metadata
        self._response_metadata = chunk.chunk["responseMetadata"]

    def construct_call_response(self) -> BedrockCallResponse:
        """Constructs the call response from a consumed BedrockStream.

        Raises:
            ValueError: if the stream has not yet been consumed.
        """
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        response = ConverseResponseTypeDef(
            output={"message": cast(MessageOutputTypeDef, self.message_param)},
            stopReason=self.finish_reasons[0] if self.finish_reasons else "end_turn",
            usage=self.metadata.get("usage", {}),
            metrics=self.metadata.get("metrics", {}),
            additionalModelResponseFields={},
            trace=self.metadata.get("trace", {}),
            ResponseMetadata=self._response_metadata,
        )
        return BedrockCallResponse(
            metadata=self.metadata,
            response=response,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



from mirascope.core import anthropic, prompt_template


@anthropic.call(
    "claude-3-5-sonnet-20240620",
    call_params={
        "max_tokens": 1024,
        "extra_headers": {"anthropic-beta": "prompt-caching-2024-07-31"},
    },
)
@prompt_template(
    """
    SYSTEM:
    You are an AI assistant tasked with analyzing literary works.
    Your goal is to provide insightful commentary on themes, characters, and writing style.

    Here is the book in it's entirety: {book}

    {:cache_control}

    USER: {query}
    """
)
def analyze_book(query: str, book: str): ...


print(analyze_book("What are the major themes?", "[FULL BOOK HERE]"))



import asyncio
from io import BytesIO

from typing import AsyncGenerator, Any

from pydub import AudioSegment
from pydub.playback import play

from mirascope.beta.openai import Realtime, record_as_stream, OpenAIRealtimeTool


def format_book(title: str, author: str) -> str:
    return f"{title} by {author}"


app = Realtime("gpt-4o-realtime-preview-2024-10-01", tools=[format_book])


@app.receiver("audio")
async def receive_audio(response: AudioSegment, context: dict[str, Any]) -> None:
    play(response)


@app.receiver("audio_transcript")
async def receive_audio_transcript(response: str, context: dict[str, Any]) -> None:
    print(f"AI(audio_transcript): {response}")


@app.sender()
async def send_audio_as_stream(
    context: dict[str, Any],
) -> AsyncGenerator[BytesIO, None]:
    print("Sending audio...")
    async for stream in record_as_stream():
        yield stream


@app.function_call(format_book)
async def recommend_book(tool: OpenAIRealtimeTool, context: dict[str, Any]) -> str:
    result = tool.call()
    print(result)
    return result


asyncio.run(app.run())



from mirascope.core import Messages, mistral


def parse_recommendation(response: mistral.MistralCallResponse) -> tuple[str, str]:
    title, author = response.content.split(" by ")
    return (title, author)


@mistral.call("mistral-large-latest", output_parser=parse_recommendation)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book. Output only Title by Author")


print(recommend_book("fantasy"))
# Output: ('"The Name of the Wind"', 'Patrick Rothfuss')



from mirascope.core import mistral
from mistralai.async_client import MistralAsyncClient


@mistral.call("mistral-large-latest", client=MistralAsyncClient())
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book"



import asyncio

from mirascope.core import Messages, bedrock


@bedrock.call(model="anthropic.claude-3-haiku-20240307-v1:0", stream=True)
async def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


async def main():
    stream = await recommend_book("fantasy")
    async for chunk, _ in stream:
        print(chunk.content, end="", flush=True)


asyncio.run(main())



from mirascope.core import BaseMessageParam, openai


@openai.call("gpt-4o-mini")
def summarize(text: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Summarize this text: {text}")]


@openai.call("gpt-4o-mini")
def translate(text: str, language: str) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user",
            content=f"Translate this text to {language}: {text}",
        )
    ]


summary = summarize("Long English text here...")
translation = translate(summary.content, "french")
print(translation.content)



"""Mirascope Base Classes."""

from . import _partial, _utils
from ._call_factory import call_factory
from ._utils import BaseType
from .call_kwargs import BaseCallKwargs
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .from_call_args import FromCallArgs
from .message_param import (
    AudioPart,
    BaseMessageParam,
    CacheControlPart,
    ImagePart,
    TextPart,
)
from .messages import Messages
from .metadata import Metadata
from .prompt import BasePrompt, metadata, prompt_template
from .response_model_config_dict import ResponseModelConfigDict
from .stream import BaseStream
from .structured_stream import BaseStructuredStream
from .tool import BaseTool, GenerateJsonSchemaNoTitles, ToolConfig
from .toolkit import BaseToolKit, toolkit_tool
from .types import AudioSegment

__all__ = [
    "AudioPart",
    "AudioSegment",
    "BaseCallKwargs",
    "BaseCallParams",
    "BaseCallResponse",
    "BaseCallResponseChunk",
    "BaseDynamicConfig",
    "BaseMessageParam",
    "BasePrompt",
    "BaseStream",
    "BaseStructuredStream",
    "BaseTool",
    "BaseToolKit",
    "BaseType",
    "CacheControlPart",
    "call_factory",
    "FromCallArgs",
    "GenerateJsonSchemaNoTitles",
    "ImagePart",
    "metadata",
    "Messages",
    "Metadata",
    "prompt_template",
    "ResponseModelConfigDict",
    "TextPart",
    "ToolConfig",
    "toolkit_tool",
    "_partial",
    "_utils",
]



"""Tests the `groq.tool` module."""

from groq.types.chat import ChatCompletionToolParam
from groq.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)

from mirascope.core.base.tool import BaseTool
from mirascope.core.groq.tool import GroqTool


def test_groq_tool() -> None:
    """Tests the `GroqTool` class."""

    class FormatBook(GroqTool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ChatCompletionMessageToolCall(
        id="id",
        function=Function(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )

    tool = FormatBook.from_tool_call(tool_call)
    assert isinstance(tool, BaseTool)
    assert isinstance(tool, GroqTool)
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"

    assert FormatBook.tool_schema() == ChatCompletionToolParam(
        type="function",
        function={
            "name": "FormatBook",
            "description": "Returns the title and author nicely formatted.",
            "parameters": {
                "properties": {
                    "title": {"type": "string"},
                    "author": {"type": "string"},
                },
                "required": ["title", "author"],
                "type": "object",
            },
        },
    )



from typing import cast

from cohere import Client
from cohere.types import Tool, ToolParameterDefinitionsValue

client = Client()


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


def identify_author(book: str) -> str:
    response = client.chat(
        model="command-r-plus",
        message=f"Who wrote {book}?",
        tools=[
            Tool(
                name="get_book_author",
                description="Returns the author of the book with the given title.",
                parameter_definitions={
                    "title": ToolParameterDefinitionsValue(
                        description=None, type="string", required=True
                    )
                },
            )
        ],
    )
    if tool_calls := response.tool_calls:
        if tool_calls[0].name == "get_book_author":
            return get_book_author(**(cast(dict[str, str], tool_calls[0].parameters)))
    return response.text


author = identify_author("The Name of the Wind")
print(author)



from mirascope.core import BaseMessageParam, anthropic
from mirascope.integrations.otel import with_hyperdx


@with_hyperdx()
@anthropic.call(model="claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book.")]


print(recommend_book("fantasy"))



import logfire
from mirascope.core import anthropic, prompt_template
from mirascope.integrations.logfire import with_logfire
from pydantic import BaseModel

logfire.configure()


class Book(BaseModel):
    title: str
    author: str


@with_logfire()
@anthropic.call("claude-3-5-sonnet-20240620", response_model=Book)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



from collections.abc import Callable

from pydantic import BaseModel, ConfigDict

from ...core.base import BaseTool


class BaseCallParams(BaseModel):
    """The parameters with which to make a call."""

    model: str
    tools: list[type[BaseTool] | Callable] | None = None

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)



from mirascope.core import BaseTool, bedrock
from pydantic import ConfigDict, Field


class GetBookAuthor(BaseTool):
    """Returns the author of the book with the given title."""

    title: str = Field(
        ...,
        description="The title of the book.",
        examples=["The Name of the Wind"],
    )

    model_config = ConfigDict(
        json_schema_extra={"examples": [{"title": "The Name of the Wind"}]}
    )

    def call(self) -> str:
        if self.title == "The Name of the Wind":
            return "Patrick Rothfuss"
        elif self.title == "Mistborn: The Final Empire":
            return "Brandon Sanderson"
        else:
            return "Unknown"


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0", tools=[GetBookAuthor])
def identify_author(book: str) -> str:
    return f"Who wrote {book}?"


response = identify_author("The Name of the Wind")
if tool := response.tool:
    print(tool.call())
else:
    print(response.content)



from mirascope.core import vertex
from tenacity import retry, stop_after_attempt, wait_exponential


@vertex.call("gemini-1.5-flash", stream=True)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
def stream():
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)


stream()



from anthropic import Anthropic

client = Anthropic()


def recommend_book(genre: str) -> str:
    message = client.messages.create(
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": f"Recommend a {genre} book"}],
        max_tokens=1024,
    )
    block = message.content[0]
    return block.text if block.type == "text" else ""


output = recommend_book("fantasy")
print(output)
